@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,7 +15,7 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module for treatment of additional (in-)equality constraints for optimization."""
+"""Optimization problems with additional (in-)equatility constraints."""
 
 from cashocs._constraints.constrained_problems import ConstrainedOptimalControlProblem
 from cashocs._constraints.constrained_problems import (
@@ -27,6 +27,6 @@ from cashocs._constraints.constraints import InequalityConstraint
 __all__ = [
     "ConstrainedOptimalControlProblem",
     "ConstrainedShapeOptimizationProblem",
-    "InequalityConstraint",
     "EqualityConstraint",
+    "InequalityConstraint",
 ]
