@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,21 +15,27 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module for UFL form manipulations.
+"""Derivation and management of weak forms.
 
-This is used to carry out form manipulations such as generating the UFL
+This module is used to carry out form manipulations such as generating the UFL
 forms for the adjoint system and for the Riesz gradient identificiation
 problems.
 """
 
 from cashocs._forms.control_form_handler import ControlFormHandler
 from cashocs._forms.form_handler import FormHandler
+from cashocs._forms.general_form_handler import AdjointFormHandler
+from cashocs._forms.general_form_handler import GeneralFormHandler
+from cashocs._forms.general_form_handler import StateFormHandler
 from cashocs._forms.shape_form_handler import ShapeFormHandler
 from cashocs._forms.shape_regularization import ShapeRegularization
 
 __all__ = [
     "ControlFormHandler",
     "FormHandler",
+    "AdjointFormHandler",
+    "GeneralFormHandler",
+    "StateFormHandler",
     "ShapeFormHandler",
     "ShapeRegularization",
 ]
