@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,8 +15,4 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module gathering all core functionality for PDE constrained optimization."""
-
-from cashocs._optimization.optimization_problem import OptimizationProblem
-
-__all__ = ["OptimizationProblem"]
+"""Optimization routines."""
