@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,23 +15,30 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module for general reduced cost functionals."""
+"""Reduced cost functionals."""
 
 from __future__ import annotations
 
 import abc
-from typing import List, Optional, Set, Tuple, TYPE_CHECKING, Union
+import ctypes
+from typing import cast, TYPE_CHECKING
 
 import fenics
 import numpy as np
-import ufl
+
+try:
+    import ufl_legacy as ufl
+except ImportError:
+    import ufl
 
 from cashocs import _exceptions
+from cashocs import _forms
 from cashocs import _utils
 
 if TYPE_CHECKING:
     from cashocs import _pde_problems
-    from cashocs import types
+    from cashocs import _typing
+    from cashocs._database import database
 
 
 class ReducedCostFunctional:
@@ -39,16 +46,19 @@ class ReducedCostFunctional:
 
     def __init__(
         self,
-        form_handler: types.FormHandler,
+        db: database.Database,
+        form_handler: _typing.FormHandler,
         state_problem: _pde_problems.StateProblem,
     ) -> None:
         """Initializes self.
 
         Args:
+            db: The database of the problem.
             form_handler: The FormHandler object for the optimization problem.
             state_problem: The StateProblem object corresponding to the state system.
 
         """
+        self.db = db
         self.form_handler = form_handler
         self.state_problem = state_problem
 
@@ -66,13 +76,13 @@ class ReducedCostFunctional:
         self.state_problem.solve()
 
         vals = [
-            functional.evaluate()
-            for functional in self.form_handler.cost_functional_list
+            functional.evaluate() for functional in self.db.form_db.cost_functional_list
         ]
         val = sum(vals)
         val += self.form_handler.cost_functional_shift
 
-        if self.form_handler.is_shape_problem:
+        if self.db.parameter_db.problem_type == "shape":
+            self.form_handler = cast(_forms.ShapeFormHandler, self.form_handler)
             val += self.form_handler.shape_regularization.compute_objective()
 
         return val
@@ -81,10 +91,6 @@ class ReducedCostFunctional:
 class Functional(abc.ABC):
     """Base class for all cost functionals."""
 
-    def __init__(self) -> None:
-        """Initialize the functional."""
-        pass
-
     @abc.abstractmethod
     def evaluate(self) -> float:
         """Evaluates the functional.
@@ -112,7 +118,7 @@ class Functional(abc.ABC):
         pass
 
     @abc.abstractmethod
-    def coefficients(self) -> Tuple[fenics.Function]:
+    def coefficients(self) -> tuple[fenics.Function]:
         """Computes the ufl coefficients which are used in the functional.
 
         Returns:
@@ -122,7 +128,7 @@ class Functional(abc.ABC):
         pass
 
     @abc.abstractmethod
-    def scale(self, scaling_factor: Union[float, int]) -> None:
+    def scale(self, scaling_factor: float | int) -> None:
         """Scales the functional by a scalar.
 
         Args:
@@ -176,17 +182,17 @@ class IntegralFunctional(Functional):
         """
         return fenics.derivative(self.form, argument, direction)
 
-    def coefficients(self) -> Tuple[fenics.Function]:
+    def coefficients(self) -> tuple[fenics.Function]:
         """Computes the ufl coefficients which are used in the functional.
 
         Returns:
             The set of used coefficients.
 
         """
-        coeffs: Tuple[fenics.Function] = self.form.coefficients()
+        coeffs: tuple[fenics.Function] = self.form.coefficients()
         return coeffs
 
-    def scale(self, scaling_factor: Union[float, int]) -> None:
+    def scale(self, scaling_factor: float | int) -> None:
         """Scales the functional by a scalar.
 
         Args:
@@ -206,26 +212,29 @@ class ScalarTrackingFunctional(Functional):
     def __init__(
         self,
         integrand: ufl.Form,
-        tracking_goal: Union[float, int],
-        weight: Union[float, int] = 1.0,
+        tracking_goal: float | int | ctypes.c_float | ctypes.c_double,
+        weight: float | int = 1.0,
     ) -> None:
         """Initializes self.
 
         Args:
             integrand: The integrand of the functional
             tracking_goal: A real number, which the integral of the integrand should
-                track
+                track. Note, that when a ctypes object is passed, the float is assumed
+                to be mutable and the tracking_goal is updated every iteration.
             weight: A real number which gives the scaling factor for this functional
 
         """
         super().__init__()
         self.integrand = integrand
         self.tracking_goal = tracking_goal
-        mesh = self.integrand.integrals()[0].ufl_domain().ufl_cargo()
-        self.integrand_value = fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-        self.weight = fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-        self.weight.vector().vec().set(weight)
-        self.weight.vector().apply("")
+        if not isinstance(self.tracking_goal, (ctypes.c_float, ctypes.c_double)):
+            self.tracking_goal_value = self.tracking_goal
+        else:
+            self.tracking_goal_value = self.tracking_goal.value
+
+        self.integrand_value = fenics.Constant(0.0)
+        self.weight = fenics.Constant(weight)
 
     def evaluate(self) -> float:
         """Evaluates the functional.
@@ -234,13 +243,15 @@ class ScalarTrackingFunctional(Functional):
             The current value of the functional.
 
         """
+        if isinstance(self.tracking_goal, (ctypes.c_float, ctypes.c_double)):
+            self.tracking_goal_value = self.tracking_goal.value
+
         scalar_integral_value = fenics.assemble(self.integrand)
-        self.integrand_value.vector().vec().set(scalar_integral_value)
-        self.integrand_value.vector().apply("")
+        self.integrand_value.assign(scalar_integral_value)
         val: float = (
-            self.weight.vector().vec()[0]
+            self.weight.values()[0]
             / 2.0
-            * pow(scalar_integral_value - self.tracking_goal, 2)
+            * pow(scalar_integral_value - self.tracking_goal_value, 2)
         )
         return val
 
@@ -257,40 +268,41 @@ class ScalarTrackingFunctional(Functional):
             A form of the resulting derivative
 
         """
+        if isinstance(self.tracking_goal, (ctypes.c_float, ctypes.c_double)):
+            self.tracking_goal_value = self.tracking_goal.value
+
         derivative = fenics.derivative(
             self.weight
-            * (self.integrand_value - fenics.Constant(self.tracking_goal))
+            * (self.integrand_value - fenics.Constant(self.tracking_goal_value))
             * self.integrand,
             argument,
             direction,
         )
         return derivative
 
-    def coefficients(self) -> Tuple[fenics.Function]:
+    def coefficients(self) -> tuple[fenics.Function]:
         """Computes the ufl coefficients which are used in the functional.
 
         Returns:
             The set of used coefficients.
 
         """
-        coeffs: Tuple[fenics.Function] = self.integrand.coefficients()
+        coeffs: tuple[fenics.Function] = self.integrand.coefficients()
         return coeffs
 
-    def scale(self, scaling_factor: Union[float, int]) -> None:
+    def scale(self, scaling_factor: float | int) -> None:
         """Scales the functional by a scalar.
 
         Args:
             scaling_factor: The scaling factor used to scale the functional
 
         """
-        self.weight.vector().vec().set(scaling_factor)
-        self.weight.vector().apply("")
+        self.weight.assign(scaling_factor)
 
     def update(self) -> None:
         """Updates the functional after solving the state equation."""
         scalar_integral_value = fenics.assemble(self.integrand)
-        self.integrand_value.vector().vec().set(scalar_integral_value)
-        self.integrand_value.vector().apply("")
+        self.integrand_value.assign(scalar_integral_value)
 
 
 class MinMaxFunctional(Functional):
@@ -299,10 +311,10 @@ class MinMaxFunctional(Functional):
     def __init__(
         self,
         integrand: ufl.Form,
-        lower_bound: Optional[Union[float, int]] = None,
-        upper_bound: Optional[Union[float, int]] = None,
-        mu: Union[float, int] = 1.0,
-        lambd: Union[float, int] = 0.0,
+        lower_bound: float | int | None = None,
+        upper_bound: float | int | None = None,
+        mu: float | int = 1.0,
+        lambd: float | int = 0.0,
     ) -> None:
         """Initializes self."""
         super().__init__()
@@ -316,14 +328,10 @@ class MinMaxFunctional(Functional):
             )
         self.lower_bound = lower_bound
         self.upper_bound = upper_bound
-        mesh = self.integrand.integrals()[0].ufl_domain().ufl_cargo()
-        self.integrand_value = fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-        self.mu = fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-        self.mu.vector().vec().set(mu)
-        self.mu.vector().apply("")
-        self.lambd = fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-        self.lambd.vector().vec().set(lambd)
-        self.lambd.vector().apply("")
+
+        self.integrand_value = fenics.Constant(0.0)
+        self.mu = fenics.Constant(mu)
+        self.lambd = fenics.Constant(lambd)
 
     def evaluate(self) -> float:
         """Evaluates the functional.
@@ -333,19 +341,18 @@ class MinMaxFunctional(Functional):
 
         """
         min_max_integrand_value = fenics.assemble(self.integrand)
-        self.integrand_value.vector().vec().set(min_max_integrand_value)
-        self.integrand_value.vector().apply("")
+        self.integrand_value.assign(min_max_integrand_value)
 
         val = 0.0
         if self.lower_bound is not None:
             val += (
                 1
-                / (2 * self.mu.vector().vec()[0])
+                / (2 * self.mu.values()[0])
                 * pow(
                     np.minimum(
                         0.0,
-                        self.lambd.vector().vec()[0]
-                        + self.mu.vector().vec()[0]
+                        self.lambd.values()[0]
+                        + self.mu.values()[0]
                         * (min_max_integrand_value - self.lower_bound),
                     ),
                     2,
@@ -354,12 +361,12 @@ class MinMaxFunctional(Functional):
         if self.upper_bound is not None:
             val += (
                 1
-                / (2 * self.mu.vector().vec()[0])
+                / (2 * self.mu.values()[0])
                 * pow(
                     np.maximum(
                         0.0,
-                        self.lambd.vector().vec()[0]
-                        + self.mu.vector().vec()[0]
+                        self.lambd.values()[0]
+                        + self.mu.values()[0]
                         * (min_max_integrand_value - self.upper_bound),
                     ),
                     2,
@@ -405,17 +412,17 @@ class MinMaxFunctional(Functional):
 
         return _utils.summation(derivative_list)
 
-    def coefficients(self) -> Tuple[fenics.Function]:
+    def coefficients(self) -> tuple[fenics.Function]:
         """Computes the ufl coefficients which are used in the functional.
 
         Returns:
             The set of used coefficients.
 
         """
-        coeffs: Tuple[fenics.Function] = self.integrand.coefficients()
+        coeffs: tuple[fenics.Function] = self.integrand.coefficients()
         return coeffs
 
-    def scale(self, scaling_factor: Union[float, int]) -> None:
+    def scale(self, scaling_factor: float | int) -> None:
         """Scales the functional by a scalar.
 
         Args:
@@ -427,8 +434,7 @@ class MinMaxFunctional(Functional):
     def update(self) -> None:
         """Updates the functional after solving the state equation."""
         min_max_integrand_value = fenics.assemble(self.integrand)
-        self.integrand_value.vector().vec().set(min_max_integrand_value)
-        self.integrand_value.vector().apply("")
+        self.integrand_value.assign(min_max_integrand_value)
 
 
 class Lagrangian:
@@ -436,8 +442,8 @@ class Lagrangian:
 
     def __init__(
         self,
-        cost_functional_list: List[types.CostFunctional],
-        state_forms: List[ufl.Form],
+        cost_functional_list: list[_typing.CostFunctional],
+        state_forms: list[ufl.Form],
     ) -> None:
         """Initializes self.
 
@@ -474,7 +480,7 @@ class Lagrangian:
         derivative = cost_functional_derivative + state_forms_derivative
         return derivative
 
-    def coefficients(self) -> Set[fenics.Function]:
+    def coefficients(self) -> set[fenics.Function]:
         """Computes the ufl coefficients which are used in the functional.
 
         Returns:
