@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,9 +15,16 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module for line searches used in optimization algorithms."""
+"""Line search algorithms."""
 
 from cashocs._optimization.line_search.armijo_line_search import ArmijoLineSearch
 from cashocs._optimization.line_search.line_search import LineSearch
+from cashocs._optimization.line_search.polynomial_line_search import (
+    PolynomialLineSearch,
+)
 
-__all__ = ["ArmijoLineSearch", "LineSearch"]
+__all__ = [
+    "ArmijoLineSearch",
+    "LineSearch",
+    "PolynomialLineSearch",
+]
