@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,30 +15,36 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Class representing an optimal control problem."""
+"""Optimal control problem."""
 
 from __future__ import annotations
 
-from typing import Dict, List, Optional, TYPE_CHECKING, Union
+from typing import Callable, TYPE_CHECKING
 
 import fenics
 import numpy as np
-import ufl
+
+try:
+    import ufl_legacy as ufl
+except ImportError:
+    import ufl
 
 from cashocs import _exceptions
 from cashocs import _forms
 from cashocs import _pde_problems
 from cashocs import _utils
+from cashocs import io
+from cashocs import log
+from cashocs import verification
 from cashocs._optimization import cost_functional
-from cashocs._optimization import line_search
+from cashocs._optimization import line_search as ls
 from cashocs._optimization import optimal_control
 from cashocs._optimization import optimization_algorithms
 from cashocs._optimization import optimization_problem
-from cashocs._optimization import verification
+from cashocs._optimization.optimal_control import box_constraints
 
 if TYPE_CHECKING:
-    from cashocs import io
-    from cashocs import types
+    from cashocs import _typing
 
 
 class OptimalControlProblem(optimization_problem.OptimizationProblem):
@@ -54,160 +60,41 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
     ``y1`` and so on).
     """
 
-    controls: List[fenics.Function]
-    control_spaces: List[fenics.FunctionSpace]
-
-    def __new__(
-        cls,
-        state_forms: Union[List[ufl.Form], ufl.Form],
-        bcs_list: Union[
-            List[List[fenics.DirichletBC]], List[fenics.DirichletBC], fenics.DirichletBC
-        ],
-        cost_functional_form: Union[
-            List[types.CostFunctional], types.CostFunctional, List[ufl.Form], ufl.Form
-        ],
-        states: Union[List[fenics.Function], fenics.Function],
-        controls: Union[List[fenics.Function], fenics.Function],
-        adjoints: Union[List[fenics.Function], fenics.Function],
-        config: Optional[io.Config] = None,
-        riesz_scalar_products: Optional[Union[List[ufl.Form], ufl.Form]] = None,
-        control_constraints: Optional[List[List[Union[float, fenics.Function]]]] = None,
-        initial_guess: Optional[List[fenics.Function]] = None,
-        ksp_options: Optional[
-            Union[types.KspOptions, List[List[Union[str, int, float]]]]
-        ] = None,
-        adjoint_ksp_options: Optional[
-            Union[types.KspOptions, List[List[Union[str, int, float]]]]
-        ] = None,
-        scalar_tracking_forms: Optional[Union[List[Dict], Dict]] = None,
-        min_max_terms: Optional[Union[List[Dict], Dict]] = None,
-        desired_weights: Optional[List[float]] = None,
-        control_bcs_list: Optional[
-            Union[
-                List[List[fenics.DirichletBC]],
-                List[fenics.DirichletBC],
-                fenics.DirichletBC,
-            ]
-        ] = None,
-    ) -> OptimalControlProblem:
-        r"""Initializes self.
-
-        Args:
-            state_forms: The weak form of the state equation (user implemented). Can be
-                either a single UFL form, or a (ordered) list of UFL forms.
-            bcs_list: The list of :py:class:`fenics.DirichletBC` objects describing
-                Dirichlet (essential) boundary conditions. If this is ``None``, then no
-                Dirichlet boundary conditions are imposed.
-            cost_functional_form: UFL form of the cost functional. Can also be a list of
-                summands of the cost functional
-            states: The state variable(s), can either be a :py:class:`fenics.Function`,
-                or a list of these.
-            controls: The control variable(s), can either be a
-                :py:class:`fenics.Function`, or a list of these.
-            adjoints: The adjoint variable(s), can either be a
-                :py:class:`fenics.Function`, or a (ordered) list of these.
-            config: The config file for the problem, generated via
-                :py:func:`cashocs.create_config`. Alternatively, this can also be
-                ``None``, in which case the default configurations are used, except for
-                the optimization algorithm. This has then to be specified in the
-                :py:meth:`solve <cashocs.OptimalControlProblem.solve>` method. The
-                default is ``None``.
-            riesz_scalar_products: The scalar products of the control space. Can either
-                be None, a single UFL form, or a (ordered) list of UFL forms. If
-                ``None``, the :math:`L^2(\Omega)` product is used (default is ``None``).
-            control_constraints: Box constraints posed on the control, ``None`` means
-                that there are none (default is ``None``). The (inner) lists should
-                contain two elements of the form ``[u_a, u_b]``, where ``u_a`` is the
-                lower, and ``u_b`` the upper bound.
-            initial_guess: List of functions that act as initial guess for the state
-                variables, should be valid input for :py:func:`fenics.assign`. Defaults
-                to ``None``, which means a zero initial guess.
-            ksp_options: A list of strings corresponding to command line options for
-                PETSc, used to solve the state systems. If this is ``None``, then the
-                direct solver mumps is used (default is ``None``).
-            adjoint_ksp_options: A list of strings corresponding to command line options
-                for PETSc, used to solve the adjoint systems. If this is ``None``, then
-                the same options as for the state systems are used (default is
-                ``None``).
-            scalar_tracking_forms: A list of dictionaries that define scalar tracking
-                type cost functionals, where an integral value should be brought to a
-                desired value. Each dict needs to have the keys ``'integrand'`` and
-                ``'tracking_goal'``. Default is ``None``, i.e., no scalar tracking terms
-                are considered.
-            min_max_terms: Additional terms for the cost functional, not to be used
-                directly.
-            desired_weights: A list of values for scaling the cost functional terms. If
-                this is supplied, the cost functional has to be given as list of
-                summands. The individual terms are then scaled, so that term `i` has the
-                magnitude of `desired_weights[i]` for the initial iteration. In case
-                that `desired_weights` is `None`, no scaling is performed. Default is
-                `None`.
-            control_bcs_list: A list of boundary conditions for the control variables.
-                This is passed analogously to ``bcs_list``. Default is ``None``.
-
-        Examples:
-            Examples how to use this class can be found in the :ref:`tutorial
-            <tutorial_index>`.
-
-        """
-        use_scaling = bool(desired_weights is not None)
-
-        if use_scaling:
-            unscaled_problem = super().__new__(cls)
-            unscaled_problem.__init__(  # type: ignore
-                state_forms,
-                bcs_list,
-                cost_functional_form,
-                states,
-                controls,
-                adjoints,
-                config=config,
-                riesz_scalar_products=riesz_scalar_products,
-                control_constraints=control_constraints,
-                initial_guess=initial_guess,
-                ksp_options=ksp_options,
-                adjoint_ksp_options=adjoint_ksp_options,
-                scalar_tracking_forms=scalar_tracking_forms,
-                min_max_terms=min_max_terms,
-                desired_weights=desired_weights,
-                control_bcs_list=control_bcs_list,
-            )
-            unscaled_problem._scale_cost_functional()  # overwrites cost functional list
-
-        return super().__new__(cls)
-
     def __init__(
         self,
-        state_forms: Union[List[ufl.Form], ufl.Form],
-        bcs_list: Union[
-            List[List[fenics.DirichletBC]], List[fenics.DirichletBC], fenics.DirichletBC
-        ],
-        cost_functional_form: Union[
-            List[types.CostFunctional], types.CostFunctional, List[ufl.Form], ufl.Form
-        ],
-        states: Union[List[fenics.Function], fenics.Function],
-        controls: Union[List[fenics.Function], fenics.Function],
-        adjoints: Union[List[fenics.Function], fenics.Function],
-        config: Optional[io.Config] = None,
-        riesz_scalar_products: Optional[Union[List[ufl.Form], ufl.Form]] = None,
-        control_constraints: Optional[List[List[Union[float, fenics.Function]]]] = None,
-        initial_guess: Optional[List[fenics.Function]] = None,
-        ksp_options: Optional[
-            Union[types.KspOptions, List[List[Union[str, int, float]]]]
-        ] = None,
-        adjoint_ksp_options: Optional[
-            Union[types.KspOptions, List[List[Union[str, int, float]]]]
-        ] = None,
-        scalar_tracking_forms: Optional[Union[List[Dict], Dict]] = None,
-        min_max_terms: Optional[Union[List[Dict], Dict]] = None,
-        desired_weights: Optional[List[float]] = None,
-        control_bcs_list: Optional[
-            Union[
-                List[List[fenics.DirichletBC]],
-                List[fenics.DirichletBC],
-                fenics.DirichletBC,
-            ]
-        ] = None,
+        state_forms: list[ufl.Form] | ufl.Form,
+        bcs_list: (
+            list[list[fenics.DirichletBC]]
+            | list[fenics.DirichletBC]
+            | fenics.DirichletBC
+        ),
+        cost_functional_form: list[_typing.CostFunctional] | _typing.CostFunctional,
+        states: list[fenics.Function] | fenics.Function,
+        controls: list[fenics.Function] | fenics.Function,
+        adjoints: list[fenics.Function] | fenics.Function,
+        config: io.Config | None = None,
+        riesz_scalar_products: list[ufl.Form] | ufl.Form | None = None,
+        control_constraints: list[list[float | fenics.Function]] | None = None,
+        initial_guess: list[fenics.Function] | None = None,
+        ksp_options: _typing.KspOption | list[_typing.KspOption] | None = None,
+        adjoint_ksp_options: _typing.KspOption | list[_typing.KspOption] | None = None,
+        gradient_ksp_options: _typing.KspOption | list[_typing.KspOption] | None = None,
+        desired_weights: list[float] | None = None,
+        control_bcs_list: (
+            list[list[fenics.DirichletBC]]
+            | list[fenics.DirichletBC]
+            | fenics.DirichletBC
+            | None
+        ) = None,
+        preconditioner_forms: list[ufl.Form] | ufl.Form | None = None,
+        pre_callback: Callable | None = None,
+        post_callback: Callable | None = None,
+        linear_solver: _utils.linalg.LinearSolver | None = None,
+        adjoint_linear_solver: _utils.linalg.LinearSolver | None = None,
+        newton_linearizations: ufl.Form | list[ufl.Form] | None = None,
+        excluded_from_time_derivative: (
+            list[int] | list[list[int]] | list[None] | None
+        ) = None,
     ) -> None:
         r"""Initializes self.
 
@@ -226,7 +113,7 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
             adjoints: The adjoint variable(s), can either be a
                 :py:class:`fenics.Function`, or a (ordered) list of these.
             config: The config file for the problem, generated via
-                :py:func:`cashocs.create_config`. Alternatively, this can also be
+                :py:func:`cashocs.load_config`. Alternatively, this can also be
                 ``None``, in which case the default configurations are used, except for
                 the optimization algorithm. This has then to be specified in the
                 :py:meth:`solve <cashocs.OptimalControlProblem.solve>` method. The
@@ -238,23 +125,20 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
                 that there are none (default is ``None``). The (inner) lists should
                 contain two elements of the form ``[u_a, u_b]``, where ``u_a`` is the
                 lower, and ``u_b`` the upper bound.
-            initial_guess: List of functions that act as initial guess for the state
+            initial_guess: list of functions that act as initial guess for the state
                 variables, should be valid input for :py:func:`fenics.assign`. Defaults
                 to ``None``, which means a zero initial guess.
-            ksp_options: A list of strings corresponding to command line options for
+            ksp_options: A list of dicts corresponding to command line options for
                 PETSc, used to solve the state systems. If this is ``None``, then the
                 direct solver mumps is used (default is ``None``).
-            adjoint_ksp_options: A list of strings corresponding to command line options
+            adjoint_ksp_options: A list of dicts corresponding to command line options
                 for PETSc, used to solve the adjoint systems. If this is ``None``, then
                 the same options as for the state systems are used (default is
                 ``None``).
-            scalar_tracking_forms: A list of dictionaries that define scalar tracking
-                type cost functionals, where an integral value should be brought to a
-                desired value. Each dict needs to have the keys ``'integrand'`` and
-                ``'tracking_goal'``. Default is ``None``, i.e., no scalar tracking terms
-                are considered.
-            min_max_terms: Additional terms for the cost functional, not to be used
-                directly.
+            gradient_ksp_options: A list of dicts corresponding to command line options
+                for PETSc, used to compute the (shape) gradient. If this is ``None``,
+                either a direct or an iterative method is used (depending on the
+                configuration, section OptimizationRoutine, key gradient_method).
             desired_weights: A list of values for scaling the cost functional terms. If
                 this is supplied, the cost functional has to be given as list of
                 summands. The individual terms are then scaled, so that term `i` has the
@@ -263,6 +147,24 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
                 `None`.
             control_bcs_list: A list of boundary conditions for the control variables.
                 This is passed analogously to ``bcs_list``. Default is ``None``.
+            preconditioner_forms: The list of forms for the preconditioner. The default
+                is `None`, so that the preconditioner matrix is the same as the system
+                matrix.
+            pre_callback: A function (without arguments) that will be called before each
+                solve of the state system
+            post_callback: A function (without arguments) that will be called after the
+                computation of the gradient.
+            linear_solver: The linear solver (KSP) which is used to solve the linear
+                systems arising from the discretized PDE.
+            adjoint_linear_solver: The linear solver (KSP) which is used to solve the
+                (linear) adjoint system.
+            newton_linearizations: A (list of) UFL forms describing which (alternative)
+                linearizations should be used for the (nonlinear) state equations when
+                solving them (with Newton's method). The default is `None`, so that the
+                Jacobian of the supplied state forms is used.
+            excluded_from_time_derivative: For each state equation, a list of indices
+                which are not part of the first order time derivative for pseudo time
+                stepping. Example: Pressure for incompressible flow. Default is None.
 
         Examples:
             Examples how to use this class can be found in the :ref:`tutorial
@@ -275,32 +177,47 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
             cost_functional_form,
             states,
             adjoints,
-            config,
-            initial_guess,
-            ksp_options,
-            adjoint_ksp_options,
-            scalar_tracking_forms,
-            min_max_terms,
-            desired_weights,
+            config=config,
+            initial_guess=initial_guess,
+            ksp_options=ksp_options,
+            adjoint_ksp_options=adjoint_ksp_options,
+            gradient_ksp_options=gradient_ksp_options,
+            desired_weights=desired_weights,
+            preconditioner_forms=preconditioner_forms,
+            pre_callback=pre_callback,
+            post_callback=post_callback,
+            linear_solver=linear_solver,
+            adjoint_linear_solver=adjoint_linear_solver,
+            newton_linearizations=newton_linearizations,
+            excluded_from_time_derivative=excluded_from_time_derivative,
         )
 
-        self.controls = _utils.enlist(controls)
-        self.control_dim = len(self.controls)
+        self.db.function_db.controls = _utils.enlist(controls)
 
-        # riesz_scalar_products
-        self.riesz_scalar_products = self._parse_riesz_scalar_products(
+        self.db.parameter_db.control_dim = len(self.db.function_db.controls)
+        self.db.function_db.control_spaces = [
+            x.function_space() for x in self.db.function_db.controls
+        ]
+        self.db.function_db.gradient = _utils.create_function_list(
+            self.db.function_db.control_spaces
+        )
+        self.db.parameter_db.problem_type = "control"
+
+        self.mesh_parametrization = None
+
+        self.riesz_scalar_products: list[ufl.Form] = self._parse_riesz_scalar_products(
             riesz_scalar_products
         )
 
         self.use_control_bcs = False
-        self.control_bcs_list: Union[List[List[fenics.DirichletBC]], List[None]]
+        self.control_bcs_list: list[list[fenics.DirichletBC]] | list[None]
         if control_bcs_list is not None:
             self.control_bcs_list_inhomogeneous = _utils.check_and_enlist_bcs(
                 control_bcs_list
             )
-            self.control_bcs_list = []  # type: ignore
+            self.control_bcs_list = []
             for list_bcs in self.control_bcs_list_inhomogeneous:
-                hom_bcs: List[fenics.DirichletBC] = [
+                hom_bcs: list[fenics.DirichletBC] = [
                     fenics.DirichletBC(bc) for bc in list_bcs
                 ]
                 for bc in hom_bcs:
@@ -309,43 +226,72 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
 
             self.use_control_bcs = True
         else:
-            self.control_bcs_list = [None] * self.control_dim
+            self.control_bcs_list = [None] * len(self.db.function_db.controls)
 
         # control_constraints
-        self.control_constraints = self._parse_control_constraints(control_constraints)
-        self._validate_control_constraints()
-
+        self.box_constraints = box_constraints.BoxConstraints(
+            self.db.function_db.controls, control_constraints
+        )
+        self.db.parameter_db.display_box_constraints = (
+            self.box_constraints.display_box_constraints
+        )
         # end overloading
 
         self.is_control_problem = True
-        self.form_handler: _forms.ControlFormHandler = _forms.ControlFormHandler(self)
-
-        self.state_spaces = self.form_handler.state_spaces
-        self.control_spaces = self.form_handler.control_spaces
-        self.adjoint_spaces = self.form_handler.adjoint_spaces
-
-        self.projected_difference = _utils.create_function_list(self.control_spaces)
-
-        self.state_problem = _pde_problems.StateProblem(
-            self.form_handler, self.initial_guess
+        self.form_handler: _forms.ControlFormHandler = _forms.ControlFormHandler(
+            self, self.db
         )
-        self.adjoint_problem = _pde_problems.AdjointProblem(
-            self.form_handler, self.state_problem
+
+        self.state_spaces = self.db.function_db.state_spaces
+        self.adjoint_spaces = self.db.function_db.adjoint_spaces
+
+        self.projected_difference = _utils.create_function_list(
+            self.db.function_db.control_spaces
         )
+
         self.gradient_problem: _pde_problems.ControlGradientProblem = (
             _pde_problems.ControlGradientProblem(
-                self.form_handler, self.state_problem, self.adjoint_problem
+                self.db, self.form_handler, self.state_problem, self.adjoint_problem
             )
         )
 
-        self.algorithm = _utils.optimization_algorithm_configuration(self.config)
-
         self.reduced_cost_functional = cost_functional.ReducedCostFunctional(
-            self.form_handler, self.state_problem
+            self.db, self.form_handler, self.state_problem
         )
 
-        self.gradient = self.gradient_problem.gradient
-        self.objective_value = 1.0
+        self.optimization_variable_abstractions = (
+            optimal_control.ControlVariableAbstractions(
+                self, self.box_constraints, self.db
+            )
+        )
+        self._silent = False
+
+        if bool(desired_weights is not None):
+            self._scale_cost_functional()
+            self.__init__(  # type: ignore
+                state_forms,
+                bcs_list,
+                cost_functional_form,
+                states,
+                controls,
+                adjoints,
+                config=config,
+                riesz_scalar_products=riesz_scalar_products,
+                control_constraints=control_constraints,
+                initial_guess=initial_guess,
+                ksp_options=ksp_options,
+                adjoint_ksp_options=adjoint_ksp_options,
+                gradient_ksp_options=gradient_ksp_options,
+                desired_weights=None,
+                control_bcs_list=control_bcs_list,
+                preconditioner_forms=preconditioner_forms,
+                pre_callback=pre_callback,
+                post_callback=post_callback,
+                linear_solver=linear_solver,
+                adjoint_linear_solver=adjoint_linear_solver,
+                newton_linearizations=newton_linearizations,
+                excluded_from_time_derivative=excluded_from_time_derivative,
+            )
 
     def _erase_pde_memory(self) -> None:
         """Resets the memory of the PDE problems so that new solutions are computed.
@@ -359,16 +305,63 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
     def _setup_control_bcs(self) -> None:
         """Sets up the boundary conditions for the control variables."""
         if self.use_control_bcs:
-            for i in range(self.control_dim):
+            for i in range(len(self.db.function_db.controls)):
                 for bc in self.control_bcs_list_inhomogeneous[i]:
-                    bc.apply(self.controls[i].vector())
+                    bc.apply(self.db.function_db.controls[i].vector())
+
+    def _setup_solver(self) -> optimization_algorithms.OptimizationAlgorithm:
+        line_search_type = self.config.get("LineSearch", "method").casefold()
+        if line_search_type == "armijo":
+            line_search: ls.LineSearch = ls.ArmijoLineSearch(self.db, self)
+        elif line_search_type == "polynomial":
+            line_search = ls.PolynomialLineSearch(self.db, self)
+        else:
+            raise _exceptions.CashocsException("This code cannot be reached.")
+
+        if self.algorithm.casefold() == "newton":
+            self.form_handler.hessian_form_handler.compute_newton_forms()
+            self.hessian_problem = _pde_problems.HessianProblem(
+                self.db,
+                self.form_handler,
+                self.general_form_handler.adjoint_form_handler,
+                self.gradient_problem,
+                self.box_constraints,
+            )
+
+        if self.algorithm.casefold() == "gradient_descent":
+            solver: optimization_algorithms.OptimizationAlgorithm = (
+                optimization_algorithms.GradientDescentMethod(
+                    self.db, self, line_search
+                )
+            )
+        elif self.algorithm.casefold() == "lbfgs":
+            solver = optimization_algorithms.LBFGSMethod(self.db, self, line_search)
+        elif self.algorithm.casefold() == "conjugate_gradient":
+            solver = optimization_algorithms.NonlinearCGMethod(
+                self.db, self, line_search
+            )
+        elif self.algorithm.casefold() == "newton":
+            solver = optimization_algorithms.NewtonMethod(self.db, self, line_search)
+        elif self.algorithm.casefold() == "none":
+            raise _exceptions.InputError(
+                "cashocs.OptimalControlProblem.solve",
+                "algorithm",
+                "You did not specify a solution algorithm in your config file. "
+                "You have to specify one in the solve method. Needs to be one of"
+                "'gradient_descent' ('gd'), 'lbfgs' ('bfgs'), 'conjugate_gradient' "
+                "('cg'), or 'newton'.",
+            )
+        else:
+            raise _exceptions.CashocsException("This code cannot be reached.")
+
+        return solver
 
     def solve(
         self,
-        algorithm: Optional[str] = None,
-        rtol: Optional[float] = None,
-        atol: Optional[float] = None,
-        max_iter: Optional[int] = None,
+        algorithm: str | None = None,
+        rtol: float | None = None,
+        atol: float | None = None,
+        max_iter: int | None = None,
     ) -> None:
         r"""Solves the optimization problem by the method specified in the config file.
 
@@ -399,66 +392,36 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
         Notes:
             If either ``rtol`` or ``atol`` are specified as arguments to the ``.solve``
             call, the termination criterion changes to:
-
             - a purely relative one (if only ``rtol`` is specified), i.e.,
 
-            .. math:: || \nabla J(u_k) || \leq \texttt{rtol} || \nabla J(u_0) ||.
+            .. math::
+                || \nabla J(u_k) || \leq \texttt{rtol} || \nabla J(u_0) ||.
 
             - a purely absolute one (if only ``atol`` is specified), i.e.,
 
-            .. math:: || \nabla J(u_K) || \leq \texttt{atol}.
+            .. math::
+                || \nabla J(u_K) || \leq \texttt{atol}.
 
             - a combined one if both ``rtol`` and ``atol`` are specified, i.e.,
 
             .. math::
-
-                || \nabla J(u_k) || \leq \texttt{atol} + \texttt{rtol}
-                || \nabla J(u_0) ||
+                || \nabla J(u_k) || \leq \texttt{atol}
+                + \texttt{rtol} || \nabla J(u_0) ||
 
         """
+        if not self._silent:
+            log.begin("Solving the optimal control problem.", level=log.INFO)
         super().solve(algorithm=algorithm, rtol=rtol, atol=atol, max_iter=max_iter)
 
         self._setup_control_bcs()
 
-        self.optimization_variable_abstractions = (
-            optimal_control.ControlVariableAbstractions(self)
-        )
-        self.line_search = line_search.ArmijoLineSearch(self)
-
-        if self.algorithm.casefold() == "newton":
-            self.form_handler.compute_newton_forms()
-
-        if self.algorithm.casefold() == "newton":
-            self.hessian_problem = _pde_problems.HessianProblem(
-                self.form_handler, self.gradient_problem
-            )
-
-        if self.algorithm.casefold() == "gradient_descent":
-            self.solver = optimization_algorithms.GradientDescentMethod(
-                self, self.line_search
-            )
-        elif self.algorithm.casefold() == "lbfgs":
-            self.solver = optimization_algorithms.LBFGSMethod(self, self.line_search)
-        elif self.algorithm.casefold() == "conjugate_gradient":
-            self.solver = optimization_algorithms.NonlinearCGMethod(
-                self, self.line_search
-            )
-        elif self.algorithm.casefold() == "newton":
-            self.solver = optimization_algorithms.NewtonMethod(self, self.line_search)
-        elif self.algorithm.casefold() == "none":
-            raise _exceptions.InputError(
-                "cashocs.OptimalControlProblem.solve",
-                "algorithm",
-                "You did not specify a solution algorithm in your config file. "
-                "You have to specify one in the solve method. Needs to be one of"
-                "'gradient_descent' ('gd'), 'lbfgs' ('bfgs'), 'conjugate_gradient' "
-                "('cg'), or 'newton'.",
-            )
-
+        self.solver = self._setup_solver()
         self.solver.run()
         self.solver.post_processing()
+        if not self._silent:
+            log.end()
 
-    def compute_gradient(self) -> List[fenics.Function]:
+    def compute_gradient(self) -> list[fenics.Function]:
         """Solves the Riesz problem to determine the gradient.
 
         This can be used for debugging, or code validation. The necessary solutions of
@@ -470,9 +433,9 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
         """
         self.gradient_problem.solve()
 
-        return self.gradient
+        return self.db.function_db.gradient
 
-    def supply_derivatives(self, derivatives: Union[ufl.Form, List[ufl.Form]]) -> None:
+    def supply_derivatives(self, derivatives: ufl.Form | list[ufl.Form]) -> None:
         """Overwrites the derivatives of the reduced cost functional w.r.t. controls.
 
         This allows users to implement their own derivatives and use cashocs as a
@@ -483,7 +446,7 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
             the control variables.
 
         """
-        mod_derivatives: List[ufl.Form]
+        mod_derivatives: list[ufl.Form]
         if isinstance(derivatives, ufl.form.Form):
             mod_derivatives = [derivatives]
         else:
@@ -500,11 +463,13 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
 
     def supply_custom_forms(
         self,
-        derivatives: Union[ufl.Form, List[ufl.Form]],
-        adjoint_forms: Union[ufl.Form, List[ufl.Form]],
-        adjoint_bcs_list: Union[
-            fenics.DirichletBC, List[fenics.DirichletBC], List[List[fenics.DirichletBC]]
-        ],
+        derivatives: ufl.Form | list[ufl.Form],
+        adjoint_forms: ufl.Form | list[ufl.Form],
+        adjoint_bcs_list: (
+            fenics.DirichletBC
+            | list[fenics.DirichletBC]
+            | list[list[fenics.DirichletBC]]
+        ),
     ) -> None:
         """Overrides both adjoint system and derivatives with user input.
 
@@ -525,9 +490,9 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
 
     def gradient_test(
         self,
-        u: Optional[List[fenics.Function]] = None,
-        h: Optional[List[fenics.Function]] = None,
-        rng: Optional[np.random.RandomState] = None,
+        u: list[fenics.Function] | None = None,
+        h: list[fenics.Function] | None = None,
+        rng: np.random.RandomState | None = None,
     ) -> float:
         """Performs a Taylor test to verify correctness of the computed gradient.
 
@@ -547,8 +512,8 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
         return verification.control_gradient_test(self, u, h, rng)
 
     def _parse_riesz_scalar_products(
-        self, riesz_scalar_products: Union[List[ufl.Form], ufl.Form]
-    ) -> List[ufl.Form]:
+        self, riesz_scalar_products: list[ufl.Form] | ufl.Form
+    ) -> list[ufl.Form]:
         """Checks, whether a given scalar product is symmetric.
 
         Args:
@@ -559,143 +524,17 @@ class OptimalControlProblem(optimization_problem.OptimizationProblem):
 
         """
         if riesz_scalar_products is None:
-            dx = fenics.Measure("dx", self.controls[0].function_space().mesh())
+            dx = ufl.Measure(
+                "dx", self.db.function_db.controls[0].function_space().mesh()
+            )
             return [
-                fenics.inner(
-                    fenics.TrialFunction(self.controls[i].function_space()),
-                    fenics.TestFunction(self.controls[i].function_space()),
+                ufl.inner(
+                    fenics.TrialFunction(self.db.function_db.control_spaces[i]),
+                    fenics.TestFunction(self.db.function_db.control_spaces[i]),
                 )
                 * dx
-                for i in range(len(self.controls))
+                for i in range(len(self.db.function_db.controls))
             ]
         else:
             self.uses_custom_scalar_product = True
             return _utils.enlist(riesz_scalar_products)
-
-    def _parse_control_constraints(
-        self,
-        control_constraints: Optional[List[List[Union[float, fenics.Function]]]],
-    ) -> List[List[fenics.Function]]:
-        """Checks, whether the given control constraints are feasible.
-
-        Args:
-            control_constraints: The control constraints.
-
-        Returns:
-            The (wrapped) list of control constraints.
-
-        """
-        temp_control_constraints: List[List[Union[fenics.Function, float]]]
-        if control_constraints is None:
-            temp_control_constraints = []
-            for control in self.controls:
-                u_a = fenics.Function(control.function_space())
-                u_a.vector().vec().set(float("-inf"))
-                u_a.vector().apply("")
-                u_b = fenics.Function(control.function_space())
-                u_b.vector().vec().set(float("inf"))
-                u_b.vector().apply("")
-                temp_control_constraints.append([u_a, u_b])
-        else:
-            temp_control_constraints = _utils.check_and_enlist_control_constraints(
-                control_constraints
-            )
-
-        # recast floats into functions for compatibility
-        formatted_control_constraints: List[List[fenics.Function]] = []
-        for idx, pair in enumerate(temp_control_constraints):
-            if isinstance(pair[0], fenics.Function):
-                lower_bound = pair[0]
-            else:
-                lower_bound = fenics.Function(self.controls[idx].function_space())
-                lower_bound.vector().vec().set(pair[0])
-                lower_bound.vector().apply("")
-
-            if isinstance(pair[1], fenics.Function):
-                upper_bound = pair[1]
-            else:
-                upper_bound = fenics.Function(self.controls[idx].function_space())
-                upper_bound.vector().vec().set(pair[1])
-                upper_bound.vector().apply("")
-
-            formatted_control_constraints.append([lower_bound, upper_bound])
-
-        return formatted_control_constraints
-
-    def _validate_control_constraints(self) -> None:
-        """Checks, whether given control constraints are valid."""
-        self.require_control_constraints = [False] * self.control_dim
-        for idx, pair in enumerate(self.control_constraints):
-            if not np.all(pair[0].vector()[:] < pair[1].vector()[:]):
-                raise _exceptions.InputError(
-                    (
-                        "cashocs._optimization.optimal_control."
-                        "optimal_control_problem.OptimalControlProblem"
-                    ),
-                    "control_constraints",
-                    (
-                        "The lower bound must always be smaller than the upper bound "
-                        "for the control_constraints."
-                    ),
-                )
-
-            if pair[0].vector().vec().max()[1] == float("-inf") and pair[
-                1
-            ].vector().vec().min()[1] == float("inf"):
-                # no control constraint for this component
-                pass
-            else:
-                self.require_control_constraints[idx] = True
-
-                control_element = self.controls[idx].ufl_element()
-                if control_element.family() == "Mixed":
-                    for j in range(control_element.value_size()):
-                        sub_elem = control_element.extract_component(j)[1]
-                        if not (
-                            sub_elem.family() == "Real"
-                            or (
-                                sub_elem.family() == "Lagrange"
-                                and sub_elem.degree() == 1
-                            )
-                            or (
-                                sub_elem.family() == "Discontinuous Lagrange"
-                                and sub_elem.degree() == 0
-                            )
-                        ):
-                            raise _exceptions.InputError(
-                                (
-                                    "cashocs._optimization.optimal_control."
-                                    "optimal_control_problem.OptimalControlProblem"
-                                ),
-                                "controls",
-                                (
-                                    "Control constraints are only implemented for "
-                                    "linear Lagrange, constant Discontinuous Lagrange, "
-                                    "and Real elements."
-                                ),
-                            )
-
-                else:
-                    if not (
-                        control_element.family() == "Real"
-                        or (
-                            control_element.family() == "Lagrange"
-                            and control_element.degree() == 1
-                        )
-                        or (
-                            control_element.family() == "Discontinuous Lagrange"
-                            and control_element.degree() == 0
-                        )
-                    ):
-                        raise _exceptions.InputError(
-                            (
-                                "cashocs._optimization.optimal_control."
-                                "optimal_control_problem.OptimalControlProblem"
-                            ),
-                            "controls",
-                            (
-                                "Control constraints are only implemented for "
-                                "linear Lagrange, constant Discontinuous Lagrange, "
-                                "and Real elements."
-                            ),
-                        )
