@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,22 +15,24 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module for managing optimization variables.
+"""Management of optimization variables.
 
 This is used to update, restore, and manipulate optimization with abstractions, so that
-the same optimization algorithms can be used for different types of problems.
+the same optimization algorithms can be used for different _typing of problems.
 """
 
 from __future__ import annotations
 
 import abc
-from typing import List, TYPE_CHECKING
+from typing import TYPE_CHECKING
 
 import fenics
+import numpy as np
 
 if TYPE_CHECKING:
+    from cashocs import _typing
     from cashocs import geometry
-    from cashocs import types
+    from cashocs._database import database
 
 
 class OptimizationVariableAbstractions(abc.ABC):
@@ -38,19 +40,24 @@ class OptimizationVariableAbstractions(abc.ABC):
 
     mesh_handler: geometry._MeshHandler  # pylint: disable=protected-access
 
-    def __init__(self, optimization_problem: types.OptimizationProblem) -> None:
+    def __init__(
+        self, optimization_problem: _typing.OptimizationProblem, db: database.Database
+    ) -> None:
         """Initializes self.
 
         Args:
             optimization_problem: The corresponding optimization problem.
+            db: The database of the problem.
 
         """
-        self.gradient = optimization_problem.gradient
+        self.db = db
+
         self.form_handler = optimization_problem.form_handler
+        self.deformation = fenics.Function(self.db.function_db.control_spaces[0])
 
     @abc.abstractmethod
     def compute_decrease_measure(
-        self, search_direction: List[fenics.Function]
+        self, search_direction: list[fenics.Function]
     ) -> float:
         """Computes the measure of decrease needed for the Armijo test.
 
@@ -70,7 +77,13 @@ class OptimizationVariableAbstractions(abc.ABC):
 
     @abc.abstractmethod
     def update_optimization_variables(
-        self, search_direction: List[fenics.Function], stepsize: float, beta: float
+        self,
+        search_direction: list[fenics.Function],
+        stepsize: float,
+        beta: float,
+        active_idx: np.ndarray | None = None,
+        constraint_gradient: np.ndarray | None = None,
+        dropped_idx: np.ndarray | None = None,
     ) -> float:
         """Updates the optimization variables based on a line search.
 
@@ -79,6 +92,13 @@ class OptimizationVariableAbstractions(abc.ABC):
             stepsize: The current (trial) stepsize.
             beta: The parameter for the line search, which "halves" the stepsize if the
                 test was not successful.
+            active_idx: The list of active indices of the working set. Only needed
+                for shape optimization with mesh quality constraints. Default is `None`.
+            constraint_gradient: The gradient of the constraints for the mesh quality.
+                Only needed for shape optimization with mesh quality constraints.
+                Default is `None`.
+            dropped_idx: The list of indicies for dropped constraints. Only needed
+                for shape optimization with mesh quality constraints. Default is `None`.
 
         Returns:
             The stepsize which was found to be acceptable.
@@ -98,7 +118,7 @@ class OptimizationVariableAbstractions(abc.ABC):
 
     @abc.abstractmethod
     def compute_a_priori_decreases(
-        self, search_direction: List[fenics.Function], stepsize: float
+        self, search_direction: list[fenics.Function], stepsize: float
     ) -> int:
         """Computes the number of times the stepsize has to be "halved" a priori.
 
@@ -124,7 +144,7 @@ class OptimizationVariableAbstractions(abc.ABC):
 
     @abc.abstractmethod
     def project_ncg_search_direction(
-        self, search_direction: List[fenics.Function]
+        self, search_direction: list[fenics.Function]
     ) -> None:
         """Restricts the search direction to the inactive set.
 
@@ -133,3 +153,54 @@ class OptimizationVariableAbstractions(abc.ABC):
 
         """
         pass
+
+    @abc.abstractmethod
+    def compute_active_sets(self) -> None:
+        """Computes the active sets of the problem."""
+        pass
+
+    def restrict_to_inactive_set(
+        self, a: list[fenics.Function], b: list[fenics.Function]
+    ) -> list[fenics.Function]:
+        """Restricts a function to the inactive set.
+
+        Note, that nothing will happen if the type of the optimization problem does not
+        support box constraints.
+
+        Args:
+            a: The function, which shall be restricted to the inactive set
+            b: The output function, which will contain the result and is overridden.
+
+        Returns:
+            The result of the restriction (overrides input b)
+
+        """
+        for j in range(len(b)):
+            if not b[j].vector().vec().equal(a[j].vector().vec()):
+                b[j].vector().vec().aypx(0.0, a[j].vector().vec())
+                b[j].vector().apply("")
+
+        return b
+
+    # pylint: disable=unused-argument
+    def restrict_to_active_set(
+        self, a: list[fenics.Function], b: list[fenics.Function]
+    ) -> list[fenics.Function]:
+        """Restricts a function to the active set.
+
+        Note, that nothing will happen if the type of the optimization problem does not
+        support box constraints.
+
+        Args:
+            a: The function, which shall be restricted to the active set
+            b: The output function, which will contain the result and is overridden.
+
+        Returns:
+            The result of the restriction (overrides input b)
+
+        """
+        for j in range(len(b)):
+            b[j].vector().vec().set(0.0)
+            b[j].vector().apply("")
+
+        return b
