@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,7 +15,7 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Methods and classes for shape optimization problems."""
+"""Shape optimization problems."""
 
 from cashocs._optimization.shape_optimization.shape_optimization_problem import (
     ShapeOptimizationProblem,
@@ -24,4 +24,4 @@ from cashocs._optimization.shape_optimization.shape_variable_abstractions import
     ShapeVariableAbstractions,
 )
 
-__all__ = ["ShapeVariableAbstractions", "ShapeOptimizationProblem"]
+__all__ = ["ShapeOptimizationProblem", "ShapeVariableAbstractions"]
