@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,7 +15,7 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Base classes for PDE problems."""
+"""PDEs associated to optimization problems."""
 
 from cashocs._pde_problems.adjoint_problem import AdjointProblem
 from cashocs._pde_problems.control_gradient_problem import ControlGradientProblem
@@ -25,10 +25,10 @@ from cashocs._pde_problems.shape_gradient_problem import ShapeGradientProblem
 from cashocs._pde_problems.state_problem import StateProblem
 
 __all__ = [
+    "PDEProblem",
     "AdjointProblem",
     "ControlGradientProblem",
     "HessianProblem",
-    "PDEProblem",
     "ShapeGradientProblem",
     "StateProblem",
 ]
