@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,7 +15,7 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Abstract implementation of a gradient problem.
+"""Gradient problem.
 
 This class is used to solve the Riesz projection to obtain the gradient of the reduced
 cost functional.
@@ -24,15 +24,18 @@ cost functional.
 from __future__ import annotations
 
 import copy
-from typing import List, TYPE_CHECKING, Union
+from typing import TYPE_CHECKING
 
 import fenics
 
+from cashocs import _forms
 from cashocs import _utils
+from cashocs import log
 from cashocs._pde_problems import pde_problem
 
 if TYPE_CHECKING:
-    from cashocs import _forms
+    from cashocs import _typing
+    from cashocs._database import database
     from cashocs._pde_problems import adjoint_problem as ap
     from cashocs._pde_problems import state_problem as sp
 
@@ -42,6 +45,7 @@ class ControlGradientProblem(pde_problem.PDEProblem):
 
     def __init__(
         self,
+        db: database.Database,
         form_handler: _forms.ControlFormHandler,
         state_problem: sp.StateProblem,
         adjoint_problem: ap.AdjointProblem,
@@ -49,18 +53,18 @@ class ControlGradientProblem(pde_problem.PDEProblem):
         """Initializes self.
 
         Args:
+            db: The database of the problem.
             form_handler: The FormHandler object of the optimization problem.
             state_problem: The StateProblem object used to solve the state equations.
             adjoint_problem: The AdjointProblem used to solve the adjoint equations.
 
         """
-        super().__init__(form_handler)
+        super().__init__(db)
 
-        self.form_handler: _forms.ControlFormHandler
+        self.form_handler = form_handler
         self.state_problem = state_problem
         self.adjoint_problem = adjoint_problem
 
-        self.gradient = self.form_handler.gradient
         self.gradient_norm_squared = 1.0
 
         gradient_tol: float = self.config.getfloat(
@@ -69,29 +73,27 @@ class ControlGradientProblem(pde_problem.PDEProblem):
 
         gradient_method: str = self.config.get("OptimizationRoutine", "gradient_method")
 
-        option: List[List[Union[str, int, float]]] = []
-        if gradient_method.casefold() == "direct":
-            option = copy.deepcopy(_utils.linalg.direct_ksp_options)
-        elif gradient_method.casefold() == "iterative":
-            option = [
-                ["ksp_type", "cg"],
-                ["pc_type", "hypre"],
-                ["pc_hypre_type", "boomeramg"],
-                ["pc_hypre_boomeramg_strong_threshold", 0.7],
-                ["ksp_rtol", gradient_tol],
-                ["ksp_atol", 1e-50],
-                ["ksp_max_it", 250],
-            ]
+        if db.parameter_db.gradient_ksp_options is not None:
+            self.riesz_ksp_options = db.parameter_db.gradient_ksp_options
+        else:
+            if gradient_method.casefold() == "direct":
+                option: _typing.KspOption = copy.deepcopy(
+                    _utils.linalg.direct_ksp_options
+                )
+            elif gradient_method.casefold() == "iterative":
+                option = copy.deepcopy(_utils.linalg.iterative_ksp_options)
+                option["ksp_rtol"] = gradient_tol
 
-        self.riesz_ksp_options = []
-        for _ in range(self.form_handler.control_dim):
-            self.riesz_ksp_options.append(option)
+            self.riesz_ksp_options = []
+            for _ in range(len(self.db.function_db.gradient)):
+                self.riesz_ksp_options.append(option)
 
         self.b_tensors = [
-            fenics.PETScVector() for _ in range(self.form_handler.control_dim)
+            fenics.PETScVector(self.db.geometry_db.mpi_comm)
+            for _ in range(len(self.db.function_db.gradient))
         ]
 
-    def solve(self) -> List[fenics.Function]:
+    def solve(self) -> list[fenics.Function]:
         """Solves the Riesz projection problem to obtain the gradient.
 
         Returns:
@@ -102,22 +104,23 @@ class ControlGradientProblem(pde_problem.PDEProblem):
         self.adjoint_problem.solve()
 
         if not self.has_solution:
-            for i in range(self.form_handler.control_dim):
+            log.begin("Computing the gradient.", level=log.DEBUG)
+            for i in range(len(self.db.function_db.gradient)):
                 self.form_handler.assemblers[i].assemble(self.b_tensors[i])
-                _utils.solve_linear_problem(
+                self.linear_solver.solve(
+                    self.db.function_db.gradient[i],
                     A=self.form_handler.riesz_projection_matrices[i],
                     b=self.b_tensors[i].vec(),
-                    x=self.gradient[i].vector().vec(),
                     ksp_options=self.riesz_ksp_options[i],
                 )
-                self.gradient[i].vector().apply("")
 
             self.has_solution = True
 
             self.gradient_norm_squared = self.form_handler.scalar_product(
-                self.gradient, self.gradient
+                self.db.function_db.gradient, self.db.function_db.gradient
             )
 
-            self.form_handler.post_hook()
+            self.db.callback.call_post()
+            log.end()
 
-        return self.gradient
+        return self.db.function_db.gradient
