@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,36 +15,48 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Base class for all PDE problems."""
+"""General PDE problem."""
 
 from __future__ import annotations
 
 import abc
-from typing import List, TYPE_CHECKING, Union
+from typing import TYPE_CHECKING
 
 import fenics
 
+from cashocs import _utils
+
 if TYPE_CHECKING:
-    from cashocs import types
+    from cashocs._database import database
 
 
 class PDEProblem(abc.ABC):
     """Base class for a PDE problem."""
 
-    def __init__(self, form_handler: types.FormHandler) -> None:
+    def __init__(
+        self,
+        db: database.Database,
+        linear_solver: _utils.linalg.LinearSolver | None = None,
+    ) -> None:
         """Initializes self.
 
         Args:
-            form_handler: The form handler for the problem.
+            db: The database of the problem
+            linear_solver: The linear solver (KSP) which is used to solve the linear
+                systems arising from the discretized PDE
 
         """
-        self.form_handler = form_handler
-        self.config = form_handler.config
+        self.db = db
 
-        self.has_solution = False
+        self.config = db.config
+        self.has_solution: bool = False
+        if linear_solver is None:
+            self.linear_solver = _utils.linalg.LinearSolver()
+        else:
+            self.linear_solver = linear_solver
 
     @abc.abstractmethod
-    def solve(self) -> Union[fenics.Function, List[fenics.Function]]:
+    def solve(self) -> fenics.Function | list[fenics.Function]:
         """Solves the PDE.
 
         Returns:
