@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,7 +15,7 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module including utility and helper functions.
+"""Utility and helper functions.
 
 This module includes utility and helper functions used in cashocs. They might also be
 interesting for users, so they are part of the public API. Includes wrappers that allow
@@ -23,7 +23,6 @@ to shorten the coding for often recurring actions.
 """
 
 from cashocs._utils.forms import bilinear_boundary_form_modification
-from cashocs._utils.forms import create_bcs_list
 from cashocs._utils.forms import create_dirichlet_bcs
 from cashocs._utils.forms import max_
 from cashocs._utils.forms import min_
@@ -32,38 +31,49 @@ from cashocs._utils.forms import multiplication
 from cashocs._utils.forms import summation
 from cashocs._utils.helpers import check_and_enlist_bcs
 from cashocs._utils.helpers import check_and_enlist_control_constraints
-from cashocs._utils.helpers import check_and_enlist_ksp_options
+from cashocs._utils.helpers import check_file_extension
 from cashocs._utils.helpers import create_function_list
 from cashocs._utils.helpers import enlist
+from cashocs._utils.helpers import get_petsc_prefixes
+from cashocs._utils.helpers import number_of_arguments
 from cashocs._utils.helpers import optimization_algorithm_configuration
-from cashocs._utils.helpers import parse_remesh
+from cashocs._utils.helpers import tag_to_int
+from cashocs._utils.interpolations import interpolate_by_angle
+from cashocs._utils.interpolations import interpolate_by_volume
+from cashocs._utils.interpolations import interpolate_levelset_function_to_cells
 from cashocs._utils.linalg import assemble_and_solve_linear
 from cashocs._utils.linalg import assemble_petsc_system
 from cashocs._utils.linalg import Interpolator
+from cashocs._utils.linalg import l2_projection
 from cashocs._utils.linalg import setup_petsc_options
 from cashocs._utils.linalg import solve_linear_problem
 from cashocs._utils.linalg import split_linear_forms
 
 __all__ = [
-    "summation",
-    "multiplication",
+    "bilinear_boundary_form_modification",
     "create_dirichlet_bcs",
     "max_",
     "min_",
     "moreau_yosida_regularization",
-    "check_and_enlist_ksp_options",
-    "check_and_enlist_control_constraints",
+    "multiplication",
+    "summation",
     "check_and_enlist_bcs",
+    "check_and_enlist_control_constraints",
+    "check_file_extension",
+    "create_function_list",
     "enlist",
-    "parse_remesh",
+    "get_petsc_prefixes",
+    "number_of_arguments",
     "optimization_algorithm_configuration",
+    "tag_to_int",
+    "interpolate_by_angle",
+    "interpolate_by_volume",
+    "interpolate_levelset_function_to_cells",
+    "assemble_and_solve_linear",
     "assemble_petsc_system",
+    "Interpolator",
+    "l2_projection",
     "setup_petsc_options",
     "solve_linear_problem",
-    "Interpolator",
-    "create_bcs_list",
-    "assemble_and_solve_linear",
     "split_linear_forms",
-    "create_function_list",
-    "bilinear_boundary_form_modification",
 ]
