@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,38 +15,48 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Mesh generation and import tools.
+"""Mesh generation, quality, and management tools.
 
-This module consists of tools for the fast generation or import of meshes into
-fenics. The :py:func:`import_mesh <cashocs.geometry.import_mesh>` function is used to
-import (converted) GMSH mesh files, and the :py:func:`regular_mesh
-<cashocs.geometry.regular_mesh>` and :py:func:`regular_box_mesh
-<cashocs.geometry.regular_box_mesh>` commands create 2D and 3D box meshes which are
-great for testing.
+This module consists of tools for the fast generation of meshes into
+fenics. The :py:func:`regular_mesh <cashocs.geometry.regular_mesh>` and
+:py:func:`regular_box_mesh <cashocs.geometry.regular_box_mesh>` commands create 2D and
+3D box meshes which are great for testing and development.
 """
 
+from cashocs.geometry import boundary_distance
+from cashocs.geometry import deformations
+from cashocs.geometry import measure
+from cashocs.geometry import mesh
+from cashocs.geometry import mesh_handler
+from cashocs.geometry import mesh_testing
+from cashocs.geometry import quality
 from cashocs.geometry.boundary_distance import compute_boundary_distance
-from cashocs.geometry.deformation_handler import DeformationHandler
+from cashocs.geometry.deformations import DeformationHandler
 from cashocs.geometry.measure import _EmptyMeasure
 from cashocs.geometry.measure import generate_measure
-from cashocs.geometry.mesh import import_mesh
 from cashocs.geometry.mesh import interval_mesh
 from cashocs.geometry.mesh import regular_box_mesh
 from cashocs.geometry.mesh import regular_mesh
 from cashocs.geometry.mesh_handler import _MeshHandler
-from cashocs.geometry.mesh_quality import compute_mesh_quality
-from cashocs.geometry.mesh_quality import MeshQuality
+from cashocs.geometry.quality import compute_mesh_quality
+from cashocs.geometry.quality import MeshQuality
 
 __all__ = [
-    "compute_mesh_quality",
+    "boundary_distance",
+    "deformations",
+    "measure",
+    "mesh",
+    "mesh_handler",
+    "mesh_testing",
+    "quality",
     "compute_boundary_distance",
     "DeformationHandler",
-    "generate_measure",
     "_EmptyMeasure",
-    "import_mesh",
-    "regular_mesh",
+    "generate_measure",
     "interval_mesh",
     "regular_box_mesh",
+    "regular_mesh",
     "_MeshHandler",
+    "compute_mesh_quality",
     "MeshQuality",
 ]
