@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,126 +15,32 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module for mesh importing and generation."""
+"""Basic mesh generation."""
 
 from __future__ import annotations
 
 import collections
-import configparser
 import functools
-import json
-import os
-import time
-from typing import Any, Callable, Dict, List, Optional, Union
+from typing import Any, Callable, TYPE_CHECKING
 
 import fenics
+from mpi4py import MPI
 import numpy as np
 from typing_extensions import Literal
-from typing_extensions import TYPE_CHECKING
 
 from cashocs import _exceptions
-from cashocs import _loggers
-from cashocs import _utils
+from cashocs import log
+from cashocs import mpi
 from cashocs.geometry import measure
-from cashocs.geometry import mesh_quality
 
 if TYPE_CHECKING:
-    from cashocs import types
-
-
-class Mesh(fenics.Mesh):
-    """A finite element mesh."""
-
-    def __init__(self, *args: Any, **kwargs: Any) -> None:
-        """See base class."""
-        super().__init__(*args, **kwargs)
-        self._config_flag = False
-
-    def _set_config_flag(self) -> None:
-        """Indicates, that the mesh has been loaded via a config file."""
-        self._config_flag = True
-
-
-def _check_imported_mesh_quality(
-    input_arg: Union[configparser.ConfigParser, str],
-    mesh: Mesh,
-    cashocs_remesh_flag: bool,
-) -> None:
-    """Checks the quality of an imported mesh.
-
-    Args:
-        input_arg: The argument used to import the mesh.
-        mesh: The finite element mesh whose quality shall be checked.
-        cashocs_remesh_flag: A flag, indicating whether remeshing is active.
-
-    """
-    if isinstance(input_arg, configparser.ConfigParser):
-        mesh_quality_tol_lower = input_arg.getfloat("MeshQuality", "tol_lower")
-        mesh_quality_tol_upper = input_arg.getfloat("MeshQuality", "tol_upper")
-
-        if mesh_quality_tol_lower > 0.9 * mesh_quality_tol_upper:
-            _loggers.warning(
-                "You are using a lower remesh tolerance (tol_lower) close to "
-                "the upper one (tol_upper). This may slow down the "
-                "optimization considerably."
-            )
-
-        mesh_quality_measure = input_arg.get("MeshQuality", "measure")
-        mesh_quality_type = input_arg.get("MeshQuality", "type")
-
-        current_mesh_quality = mesh_quality.compute_mesh_quality(
-            mesh, mesh_quality_type, mesh_quality_measure
-        )
-
-        if not cashocs_remesh_flag:
-            if current_mesh_quality < mesh_quality_tol_lower:
-                raise _exceptions.InputError(
-                    "cashocs.geometry.import_mesh",
-                    "input_arg",
-                    "The quality of the mesh file you have specified is not "
-                    "sufficient for evaluating the cost functional.\n"
-                    f"It currently is {current_mesh_quality:.3e} but has to "
-                    f"be at least {mesh_quality_tol_lower:.3e}.",
-                )
-
-            if current_mesh_quality < mesh_quality_tol_upper:
-                raise _exceptions.InputError(
-                    "cashocs.geometry.import_mesh",
-                    "input_arg",
-                    "The quality of the mesh file you have specified is not "
-                    "sufficient for computing the shape gradient.\n "
-                    + f"It currently is {current_mesh_quality:.3e} but has to "
-                    f"be at least {mesh_quality_tol_lower:.3e}.",
-                )
-
-        else:
-            if current_mesh_quality < mesh_quality_tol_lower:
-                raise _exceptions.InputError(
-                    "cashocs.geometry.import_mesh",
-                    "input_arg",
-                    "Remeshing failed.\n"
-                    "The quality of the mesh file generated through remeshing is "
-                    "not sufficient for evaluating the cost functional.\n"
-                    + f"It currently is {current_mesh_quality:.3e} but has to "
-                    f"be at least {mesh_quality_tol_lower:.3e}.",
-                )
-
-            if current_mesh_quality < mesh_quality_tol_upper:
-                raise _exceptions.InputError(
-                    "cashocs.geometry.import_mesh",
-                    "input_arg",
-                    "Remeshing failed.\n"
-                    "The quality of the mesh file generated through remeshing "
-                    "is not sufficient for computing the shape gradient.\n "
-                    + f"It currently is {current_mesh_quality:.3e} but has to "
-                    f"be at least {mesh_quality_tol_upper:.3e}.",
-                )
+    from cashocs import _typing
 
 
 def _get_mesh_stats(
-    mode: Literal["import", "generate"] = "import"
-) -> Callable[..., Callable[..., types.MeshTuple]]:
-    """A decorator for mesh importing / generating function which prints stats.
+    mode: Literal["import", "generate"],
+) -> Callable[..., Callable[..., _typing.MeshTuple]]:
+    """A decorator for mesh importing / generating function which logs stats.
 
     Args:
         mode: A string indicating whether the mesh is being generated or imported.
@@ -145,8 +51,8 @@ def _get_mesh_stats(
     """
 
     def decorator_stats(
-        func: Callable[..., types.MeshTuple]
-    ) -> Callable[..., types.MeshTuple]:
+        func: Callable[..., _typing.MeshTuple],
+    ) -> Callable[..., _typing.MeshTuple]:
         """A decorator for a mesh generating function.
 
         Args:
@@ -158,7 +64,7 @@ def _get_mesh_stats(
         """
 
         @functools.wraps(func)
-        def wrapper_stats(*args: Any, **kwargs: Any) -> types.MeshTuple:
+        def wrapper_stats(*args: Any, **kwargs: Any) -> _typing.MeshTuple:
             """Wrapper function for mesh generating functions.
 
             Args:
@@ -169,27 +75,34 @@ def _get_mesh_stats(
                 The wrapped function.
 
             """
+            comm = None
+            if "comm" in kwargs.keys():  # pylint: disable=consider-iterating-dictionary
+                comm = kwargs["comm"]
+            else:
+                for arg in args:
+                    if isinstance(arg, MPI.Comm):
+                        comm = arg
+
+            if comm is None:
+                comm = mpi.COMM_WORLD
+
             word = "importing" if mode.casefold() == "import" else "generating"
             worded = "imported" if mode.casefold() == "import" else "generated"
-            mpi_size = fenics.MPI.size(fenics.MPI.comm_world)
-            start_time = time.time()
-            _loggers.info(f"{word.capitalize()} mesh.")
+            mpi_size = comm.size
+            log.begin(f"{word.capitalize()} mesh.", level=log.INFO)
 
             value = func(*args, **kwargs)
             dim = value[0].geometry().dim()
 
-            end_time = time.time()
-            _loggers.info(
-                f"Done {word} mesh. Elapsed time: {end_time - start_time:.2f} s."
-            )
-            _loggers.info(
+            log.info(
                 f"Successfully {worded} {dim}-dimensional mesh on {mpi_size} CPU(s)."
             )
-            _loggers.info(
+            log.info(
                 f"Mesh contains {value[0].num_entities_global(0):,} vertices and "
                 f"{value[0].num_entities_global(dim):,} cells of type "
-                f"{value[0].ufl_cell().cellname()}.\n"
+                f"{value[0].ufl_cell().cellname()}."
             )
+            log.end()
             return value
 
         return wrapper_stats
@@ -197,127 +110,14 @@ def _get_mesh_stats(
     return decorator_stats
 
 
-@_get_mesh_stats(mode="import")
-def import_mesh(input_arg: Union[str, configparser.ConfigParser]) -> types.MeshTuple:
-    """Imports a mesh file for use with cashocs / FEniCS.
-
-    This function imports a mesh file that was generated by GMSH and converted to
-    .xdmf with the command line function :ref:`cashocs-convert <cashocs_convert>`.
-    If there are Physical quantities specified in the GMSH file, these are imported
-    to the subdomains and boundaries output of this function and can also be directly
-    accessed via the measures, e.g., with ``dx(1)``, ``ds(1)``, etc.
-
-    Args:
-        input_arg: This is either a string, in which case it corresponds to the location
-            of the mesh file in .xdmf file format, or a config file that
-            has this path stored in its settings, under the section Mesh, as
-            parameter ``mesh_file``.
-
-    Returns:
-        A tuple (mesh, subdomains, boundaries, dx, ds, dS), where mesh is the imported
-        FEM mesh, subdomains is a mesh function for the subdomains, boundaries is a mesh
-        function for the boundaries, dx is a volume measure, ds is a surface measure,
-        and dS is a measure for the interior facets.
-
-    Notes:
-        In case the boundaries in the Gmsh .msh file are not only marked with numbers
-        (as physical groups), but also with names (i.e. strings), these strings can be
-        used with the integration measures ``dx`` and ``ds`` returned by this method.
-        E.g., if one specified the following in a 2D Gmsh .geo file ::
-
-            Physical Surface("domain", 1) = {i,j,k};
-
-        where i,j,k are representative for some integers, then this can be used in the
-        measure ``dx`` (as we are 2D) as follows. The command ::
-
-            dx(1)
-
-        is completely equivalent to ::
-
-           dx("domain")
-
-        and both can be used interchangeably.
-
-    """
-    cashocs_remesh_flag, temp_dir = _utils.parse_remesh()
-
-    # Check for the file format
-    mesh_file: str = ""
-    if isinstance(input_arg, str):
-        mesh_file = input_arg
-    elif isinstance(input_arg, configparser.ConfigParser):
-        if not cashocs_remesh_flag:
-            mesh_file = input_arg.get("Mesh", "mesh_file")
-        else:
-            with open(f"{temp_dir}/temp_dict.json", "r", encoding="utf-8") as file:
-                temp_dict: Dict = json.load(file)
-            mesh_file = temp_dict["mesh_file"]
-
-    file_string = mesh_file[:-5]
-
-    mesh = Mesh()
-    xdmf_file = fenics.XDMFFile(mesh.mpi_comm(), mesh_file)
-    xdmf_file.read(mesh)
-    xdmf_file.close()
-
-    subdomains_mvc = fenics.MeshValueCollection(
-        "size_t", mesh, mesh.geometric_dimension()
-    )
-    boundaries_mvc = fenics.MeshValueCollection(
-        "size_t", mesh, mesh.geometric_dimension() - 1
-    )
-
-    if os.path.isfile(f"{file_string}_subdomains.xdmf"):
-        xdmf_subdomains = fenics.XDMFFile(
-            mesh.mpi_comm(), f"{file_string}_subdomains.xdmf"
-        )
-        xdmf_subdomains.read(subdomains_mvc, "subdomains")
-        xdmf_subdomains.close()
-    if os.path.isfile(f"{file_string}_boundaries.xdmf"):
-        xdmf_boundaries = fenics.XDMFFile(
-            mesh.mpi_comm(), f"{file_string}_boundaries.xdmf"
-        )
-        xdmf_boundaries.read(boundaries_mvc, "boundaries")
-        xdmf_boundaries.close()
-
-    physical_groups: Optional[Dict[str, Dict[str, int]]] = None
-    if os.path.isfile(f"{file_string}_physical_groups.json"):
-        with open(f"{file_string}_physical_groups.json", "r", encoding="utf-8") as file:
-            physical_groups = json.load(file)
-
-    subdomains = fenics.MeshFunction("size_t", mesh, subdomains_mvc)
-    boundaries = fenics.MeshFunction("size_t", mesh, boundaries_mvc)
-
-    dx = measure.NamedMeasure(
-        "dx", domain=mesh, subdomain_data=subdomains, physical_groups=physical_groups
-    )
-    ds = measure.NamedMeasure(
-        "ds", domain=mesh, subdomain_data=boundaries, physical_groups=physical_groups
-    )
-    d_interior_facet = measure.NamedMeasure(
-        "dS", domain=mesh, subdomain_data=boundaries, physical_groups=physical_groups
-    )
-
-    # Add an attribute to the mesh to show with what procedure it was generated
-    # pylint: disable=protected-access
-    mesh._set_config_flag()
-    # Add the physical groups to the mesh in case they are present
-    if physical_groups is not None:
-        mesh.physical_groups = physical_groups
-
-    # Check the mesh quality of the imported mesh in case a config file is passed
-    _check_imported_mesh_quality(input_arg, mesh, cashocs_remesh_flag)
-
-    return mesh, subdomains, boundaries, dx, ds, d_interior_facet
-
-
-@_get_mesh_stats(mode="generate")
+@_get_mesh_stats("generate")
 def interval_mesh(
     n: int = 10,
     start: float = 0.0,
     end: float = 1.0,
-    partitions: Optional[List[float]] = None,
-) -> types.MeshTuple:
+    partitions: list[float] | None = None,
+    comm: MPI.Comm | None = None,
+) -> _typing.MeshTuple:
     r"""Creates an 1D interval mesh starting at x=0 to x=length.
 
     This function creates a uniform mesh of a 1D interval, starting at the ``start`` and
@@ -335,6 +135,7 @@ def interval_mesh(
         partitions: Points in the interval at which a partition in subdomains should be
             made. The resulting volume measure is sorted ascendingly according to the
             sub-intervals defined in partitions (starting at 1). Defaults to ``None``.
+        comm: MPI communicator that is to be used for creating the mesh.
 
     Returns:
         A tuple (mesh, subdomains, boundaries, dx, ds, dS), where mesh is the imported
@@ -359,7 +160,10 @@ def interval_mesh(
     n = int(n)
     dim = 1
 
-    mesh = fenics.IntervalMesh(n, start, end)
+    if comm is None:
+        comm = mpi.COMM_WORLD
+
+    mesh = fenics.IntervalMesh(comm, n, start, end)
 
     subdomains = fenics.MeshFunction("size_t", mesh, dim=dim)
     boundaries = fenics.MeshFunction("size_t", mesh, dim=dim - 1)
@@ -383,9 +187,10 @@ def interval_mesh(
             end_point = padded_partitions[i + 1]
 
             part = fenics.CompiledSubDomain(
-                "x[0] >= start_point && x[0] <= end_point",
+                "x[0] >= start_point - eps && x[0] <= end_point + eps",
                 start_point=start_point,
                 end_point=end_point,
+                eps=fenics.DOLFIN_EPS,
             )
             part.mark(subdomains, i + 1)
 
@@ -396,14 +201,15 @@ def interval_mesh(
     return mesh, subdomains, boundaries, dx, ds, d_interior_facet
 
 
-@_get_mesh_stats(mode="generate")
+@_get_mesh_stats("generate")
 def regular_mesh(
     n: int = 10,
     length_x: float = 1.0,
     length_y: float = 1.0,
-    length_z: Optional[float] = None,
+    length_z: float | None = None,
     diagonal: Literal["left", "right", "left/right", "right/left", "crossed"] = "right",
-) -> types.MeshTuple:
+    comm: MPI.Comm | None = None,
+) -> _typing.MeshTuple:
     r"""Creates a mesh corresponding to a rectangle or cube.
 
     This function creates a uniform mesh of either a rectangle or a cube, starting at
@@ -440,6 +246,7 @@ def regular_mesh(
         diagonal: This defines the type of diagonal used to create the box mesh in 2D.
             This can be one of ``"right"``, ``"left"``, ``"left/right"``,
             ``"right/left"`` or ``"crossed"``.
+        comm: MPI communicator that is to be used for creating the mesh.
 
     Returns:
         A tuple (mesh, subdomains, boundaries, dx, ds, dS), where mesh is the imported
@@ -465,6 +272,9 @@ def regular_mesh(
 
     n = int(n)
 
+    if comm is None:
+        comm = mpi.COMM_WORLD
+
     if length_z is None:
         sizes = [length_x, length_y]
         dim = 2
@@ -477,6 +287,7 @@ def regular_mesh(
 
     if length_z is None:
         mesh = fenics.RectangleMesh(
+            comm,
             fenics.Point(0, 0),
             fenics.Point(sizes),
             num_points[0],
@@ -485,6 +296,7 @@ def regular_mesh(
         )
     else:
         mesh = fenics.BoxMesh(
+            comm,
             fenics.Point(0, 0, 0),
             fenics.Point(sizes),
             num_points[0],
@@ -532,17 +344,18 @@ def regular_mesh(
     return mesh, subdomains, boundaries, dx, ds, d_interior_facet
 
 
-@_get_mesh_stats(mode="generate")
+@_get_mesh_stats("generate")
 def regular_box_mesh(
     n: int = 10,
     start_x: float = 0.0,
     start_y: float = 0.0,
-    start_z: Optional[float] = None,
+    start_z: float | None = None,
     end_x: float = 1.0,
     end_y: float = 1.0,
-    end_z: Optional[float] = None,
+    end_z: float | None = None,
     diagonal: Literal["right", "left", "left/right", "right/left", "crossed"] = "right",
-) -> types.MeshTuple:
+    comm: MPI.Comm | None = None,
+) -> _typing.MeshTuple:
     r"""Creates a mesh corresponding to a rectangle or cube.
 
     This function creates a uniform mesh of either a rectangle
@@ -584,6 +397,7 @@ def regular_box_mesh(
         diagonal: This defines the type of diagonal used to create the box mesh in 2D.
             This can be one of ``"right"``, ``"left"``, ``"left/right"``,
             ``"right/left"`` or ``"crossed"``.
+        comm: MPI communicator that is to be used for creating the mesh.
 
     Returns:
         A tuple (mesh, subdomains, boundaries, dx, ds, dS), where mesh is the imported
@@ -597,6 +411,9 @@ def regular_box_mesh(
     dim = 2
     sizes = [1.0, 1.0]
 
+    if comm is None:
+        comm = mpi.COMM_WORLD
+
     if start_z is None and end_z is None:
         lx = end_x - start_x
         ly = end_y - start_y
@@ -626,6 +443,7 @@ def regular_box_mesh(
 
     if start_z is None:
         mesh = fenics.RectangleMesh(
+            comm,
             fenics.Point(start_x, start_y),
             fenics.Point(end_x, end_y),
             num_points[0],
@@ -634,6 +452,7 @@ def regular_box_mesh(
         )
     else:
         mesh = fenics.BoxMesh(
+            comm,
             fenics.Point(start_x, start_y, start_z),
             fenics.Point(end_x, end_y, end_z),
             num_points[0],
@@ -679,7 +498,7 @@ def regular_box_mesh(
     return mesh, subdomains, boundaries, dx, ds, d_interior_facet
 
 
-def _check_sizes(sizes: List[float]) -> None:
+def _check_sizes(sizes: list[float]) -> None:
     for size in sizes:
         if size <= 0:
             raise _exceptions.InputError(
