@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,43 +15,23 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Module for managing config files."""
+"""Management of configuration files."""
 
 from __future__ import annotations
 
 from configparser import ConfigParser
 import json
-from pathlib import Path
-from typing import Any, Dict, List, Optional
+import pathlib
+from typing import Any
 
 from cashocs import _exceptions
-from cashocs import _loggers
 
+try:
+    import cashocs_extensions  # pylint: disable=unused-import
 
-# deprecated
-def create_config(path: str) -> ConfigParser:  # pragma: no cover
-    """Loads a config object from a config file.
-
-    Loads the config from a .ini file via the configparser package.
-
-    Args:
-        path: The path to the .ini file storing the configuration.
-
-    Returns:
-        The output config file, which includes the path to the .ini file.
-
-    .. deprecated:: 1.1.0
-        This is replaced by :py:func:`load_config <cashocs.load_config>`
-        and will be removed in the future.
-
-    """
-    _loggers.warning(
-        "DEPRECATION WARNING: cashocs.create_config is replaced by cashocs.load_config "
-        "and will be removed in the future."
-    )
-    config = load_config(path)
-
-    return config
+    has_cashocs_extensions = True
+except ImportError:
+    has_cashocs_extensions = False
 
 
 def load_config(path: str) -> ConfigParser:
@@ -82,7 +62,12 @@ def _check_for_config_list(string: str) -> bool:
     result = False
 
     for char in string:
-        if not (char.isdigit() or char.isspace() or char in ["[", "]", ".", ",", "-"]):
+        if not (
+            char.isdigit()
+            or char.isalpha()
+            or char.isspace()
+            or char in ["[", "]", ".", ",", "-", '"', "'", "_"]
+        ):
             return result
 
     if string[0] != "[":
@@ -90,15 +75,13 @@ def _check_for_config_list(string: str) -> bool:
     if string[-1] != "]":
         return result
 
-    result = True
-
-    return result
+    return True
 
 
 class Config(ConfigParser):
     """Class for handling the config in cashocs."""
 
-    def __init__(self, config_file: Optional[str] = None) -> None:
+    def __init__(self, config_file: str | None = None) -> None:
         """Initializes self.
 
         Args:
@@ -106,15 +89,10 @@ class Config(ConfigParser):
 
         """
         super().__init__()
-        self.config_errors: List[str] = []
+        self.config_errors: list[str] = []
 
-        self.config_scheme: Dict[str, Dict[str, Dict[str, Any]]] = {
+        self.config_scheme: dict[str, dict[str, dict[str, Any]]] = {
             "Mesh": {
-                "mesh_file": {
-                    "type": "str",
-                    "attributes": ["file"],
-                    "file_extension": "xdmf",
-                },
                 "gmsh_file": {
                     "type": "str",
                     "attributes": ["file"],
@@ -176,6 +154,13 @@ class Config(ConfigParser):
                 "picard_verbose": {
                     "type": "bool",
                 },
+                "backend": {
+                    "type": "str",
+                    "possible_options": ["cashocs", "petsc"],
+                },
+                "use_adjoint_linearizations": {
+                    "type": "bool",
+                },
             },
             "OptimizationRoutine": {
                 "algorithm": {
@@ -190,34 +175,23 @@ class Config(ConfigParser):
                         "nonlinear_conjugate_gradient",
                         "conjugate_gradient",
                         "newton",
+                        "sphere_combination",
+                        "convex_combination",
                         "none",
                     ],
                 },
                 "rtol": {
                     "type": "float",
-                    "attributes": ["less_than_one", "positive"],
+                    "attributes": ["less_than_one", "non_negative"],
                 },
                 "atol": {
                     "type": "float",
                     "attributes": ["non_negative"],
                 },
-                "maximum_iterations": {
+                "max_iter": {
                     "type": "int",
                     "attributes": ["non_negative"],
                 },
-                "initial_stepsize": {
-                    "type": "float",
-                    "attributes": ["positive"],
-                },
-                "safeguard_stepsize": {"type": "bool"},
-                "epsilon_armijo": {
-                    "type": "float",
-                    "attributes": ["positive", "less_than_one"],
-                },
-                "beta_armijo": {
-                    "type": "float",
-                    "attributes": ["positive", "larger_than_one"],
-                },
                 "gradient_method": {
                     "type": "str",
                     "possible_options": ["direct", "iterative"],
@@ -230,6 +204,41 @@ class Config(ConfigParser):
                     "type": "bool",
                 },
             },
+            "LineSearch": {
+                "method": {
+                    "type": "str",
+                    "possible_options": ["armijo", "polynomial"],
+                },
+                "initial_stepsize": {
+                    "type": "float",
+                    "attributes": ["positive"],
+                },
+                "epsilon_armijo": {
+                    "type": "float",
+                    "attributes": ["positive", "less_than_one"],
+                },
+                "beta_armijo": {
+                    "type": "float",
+                    "attributes": ["positive", "larger_than_one"],
+                },
+                "safeguard_stepsize": {"type": "bool"},
+                "polynomial_model": {
+                    "type": "str",
+                    "possible_options": ["cubic", "quadratic"],
+                },
+                "factor_high": {
+                    "type": "float",
+                    "attributes": ["less_than_one", "positive"],
+                    "larger_than": ("LineSearch", "factor_low"),
+                },
+                "factor_low": {
+                    "type": "float",
+                    "attributes": ["less_than_one", "positive"],
+                },
+                "fail_if_not_converged": {
+                    "type": "bool",
+                },
+            },
             "AlgoLBFGS": {
                 "bfgs_memory_size": {
                     "type": "int",
@@ -238,6 +247,13 @@ class Config(ConfigParser):
                 "use_bfgs_scaling": {
                     "type": "bool",
                 },
+                "bfgs_periodic_restart": {
+                    "type": "int",
+                    "attributes": ["non_negative"],
+                },
+                "damped": {
+                    "type": "bool",
+                },
             },
             "AlgoCG": {
                 "cg_method": {
@@ -293,6 +309,9 @@ class Config(ConfigParser):
                 "shape_bdry_fix_z": {
                     "type": "list",
                 },
+                "fixed_dimensions": {
+                    "type": "list",
+                },
                 "use_pull_back": {
                     "type": "bool",
                 },
@@ -320,6 +339,10 @@ class Config(ConfigParser):
                 "update_inhomogeneous": {
                     "type": "bool",
                 },
+                "inhomogeneous_exponent": {
+                    "type": "float",
+                    "attributes": ["non_negative"],
+                },
                 "use_distance_mu": {
                     "type": "bool",
                 },
@@ -343,6 +366,13 @@ class Config(ConfigParser):
                 "boundaries_dist": {
                     "type": "list",
                 },
+                "distance_method": {
+                    "type": "str",
+                    "possible_options": [
+                        "eikonal",
+                        "poisson",
+                    ],
+                },
                 "smooth_mu": {
                     "type": "bool",
                 },
@@ -357,12 +387,22 @@ class Config(ConfigParser):
                     "type": "float",
                     "attributes": ["non_negative", "less_than_one"],
                 },
-                "fixed_dimensions": {
-                    "type": "list",
-                },
                 "degree_estimation": {
                     "type": "bool",
                 },
+                "global_deformation": {
+                    "type": "bool",
+                },
+                "test_for_intersections": {
+                    "type": "bool",
+                },
+                "reextend_from_boundary": {
+                    "type": "bool",
+                },
+                "reextension_mode": {
+                    "type": "str",
+                    "possible_options": ["surface", "normal"],
+                },
             },
             "Regularization": {
                 "factor_volume": {
@@ -401,9 +441,6 @@ class Config(ConfigParser):
                 "use_initial_barycenter": {
                     "type": "bool",
                 },
-                "measure_hole": {
-                    "type": "bool",
-                },
                 "x_start": {
                     "type": "float",
                 },
@@ -458,7 +495,49 @@ class Config(ConfigParser):
                 },
                 "type": {
                     "type": "str",
-                    "possible_options": ["min", "avg", "minimum", "average"],
+                    "possible_options": [
+                        "min",
+                        "avg",
+                        "q",
+                        "minimum",
+                        "average",
+                        "quantile",
+                    ],
+                },
+                "quantile": {
+                    "type": "float",
+                    "attributes": ["non_negative", "less_than_one"],
+                },
+                "remesh_iter": {
+                    "type": "int",
+                    "attributes": ["non_negative"],
+                },
+            },
+            "TopologyOptimization": {
+                "angle_tol": {
+                    "type": "float",
+                    "attributes": ["positive"],
+                },
+                "interpolation_scheme": {
+                    "type": "str",
+                    "possible_options": ["angle", "volume"],
+                },
+                "normalize_topological_derivative": {
+                    "type": "bool",
+                },
+                "re_normalize_levelset": {
+                    "type": "bool",
+                },
+                "topological_derivative_is_identical": {
+                    "type": "bool",
+                },
+                "tol_bisection": {
+                    "type": "float",
+                    "attributes": ["non_negative"],
+                },
+                "max_iter_bisection": {
+                    "type": "int",
+                    "attributes": ["non_negative"],
                 },
             },
             "Output": {
@@ -471,13 +550,13 @@ class Config(ConfigParser):
                 "save_txt": {
                     "type": "bool",
                 },
-                "save_pvd": {
+                "save_state": {
                     "type": "bool",
                 },
-                "save_pvd_adjoint": {
+                "save_adjoint": {
                     "type": "bool",
                 },
-                "save_pvd_gradient": {
+                "save_gradient": {
                     "type": "bool",
                 },
                 "save_mesh": {
@@ -487,6 +566,10 @@ class Config(ConfigParser):
                 "result_dir": {
                     "type": "str",
                 },
+                "precision": {
+                    "type": "int",
+                    "attributes": ["positive"],
+                },
                 "time_suffix": {
                     "type": "bool",
                 },
@@ -519,31 +602,38 @@ picard_rtol = 1e-10
 picard_atol = 1e-12
 picard_iter = 50
 picard_verbose = False
+backend = cashocs
+use_adjoint_linearizations = False
 
 [OptimizationRoutine]
 algorithm = none
 rtol = 1e-3
 atol = 0.0
-maximum_iterations = 100
+max_iter = 100
 soft_exit = False
+gradient_tol = 1e-9
+gradient_method = direct
+
+[LineSearch]
+method = armijo
 epsilon_armijo = 1e-4
 beta_armijo = 2.0
 initial_stepsize = 1.0
 safeguard_stepsize = True
-gradient_tol = 1e-9
-gradient_method = direct
+polynomial_model = cubic
+factor_high = 0.5
+factor_low = 0.1
+fail_if_not_converged = False
 
 [ShapeGradient]
 lambda_lame = 0.0
 damping_factor = 0.0
-inhomogeneous = False
 mu_def = 1.0
 mu_fix = 1.0
 use_sqrt_mu = False
 use_p_laplacian = False
 p_laplacian_power = 2
 p_laplacian_stabilization = 0.0
-degree_estimation = True
 use_pull_back = True
 use_distance_mu = False
 mu_min = 1.0
@@ -551,14 +641,22 @@ mu_max = 1.0
 dist_min = 1.0
 dist_max = 1.0
 boundaries_dist = []
+distance_method = eikonal
 smooth_mu = False
+inhomogeneous = False
 update_inhomogeneous = False
+inhomogeneous_exponent = 1.0
 fixed_dimensions = []
 shape_bdry_def = []
 shape_bdry_fix = []
 shape_bdry_fix_x = []
 shape_bdry_fix_y = []
 shape_bdry_fix_z = []
+degree_estimation = True
+global_deformation = False
+test_for_intersections = True
+reextend_from_boundary = False
+reextension_mode = surface
 
 [Regularization]
 factor_volume = 0.0
@@ -572,7 +670,6 @@ factor_barycenter = 0.0
 target_barycenter = [0.0, 0.0, 0.0]
 use_initial_barycenter = False
 use_relative_scaling = False
-measure_hole = False
 x_start = 0.0
 x_end = 1.0
 y_start = 0.0
@@ -589,6 +686,8 @@ inner_newton_atol = 0.0
 [AlgoLBFGS]
 bfgs_memory_size = 5
 use_bfgs_scaling = True
+bfgs_periodic_restart = 0
+damped = False
 
 [AlgoCG]
 cg_method = DY
@@ -602,18 +701,30 @@ tol_lower = 0.0
 tol_upper = 1e-15
 measure = skewness
 type = min
+quantile = 0.0
 volume_change = inf
 angle_change = inf
+remesh_iter = 0
+
+[TopologyOptimization]
+angle_tol = 1.0
+interpolation_scheme = volume
+normalize_topological_derivative = False
+re_normalize_levelset = False
+topological_derivative_is_identical = False
+tol_bisection = 1e-4
+max_iter_bisection = 100
 
 [Output]
 save_results = True
-verbose = True
-save_txt = True
-save_pvd = False
-save_pvd_adjoint = False
-save_pvd_gradient = False
+verbose = False
+save_txt = False
+save_state = False
+save_adjoint = False
+save_gradient = False
 save_mesh = False
 result_dir = ./results
+precision = 3
 time_suffix = False
 
 [Debug]
@@ -623,10 +734,23 @@ restart = False
 
         self.read_string(self.default_config_str)
 
-        if config_file is not None:
-            self.read(config_file)
+        if has_cashocs_extensions:
+            self.config_scheme.update(cashocs_extensions.config.config_scheme)
+            self.read_string(cashocs_extensions.config.default_config_str)
 
-    def getlist(self, section: str, option: str, **kwargs: Any) -> List:
+        if config_file is not None:
+            file = pathlib.Path(config_file)
+            if file.is_file():
+                self.read(config_file)
+            else:
+                raise _exceptions.InputError(
+                    "cashocs.Config",
+                    "config_file",
+                    f"Could not find the specified config file {config_file}. "
+                    "Please supply a path to an existing configuration file.",
+                )
+
+    def getlist(self, section: str, option: str, **kwargs: Any) -> list:
         """Extracts a list from a config file.
 
         Args:
@@ -642,7 +766,7 @@ restart = False
         if (
             self.config_scheme[section][option]["type"] == "list"
         ) and _check_for_config_list(self.get(section, option)):
-            py_list: List = json.loads(self.get(section, option, **kwargs))
+            py_list: list = json.loads(self.get(section, option, **kwargs))
             return py_list
         else:
             raise _exceptions.InputError(
@@ -806,7 +930,7 @@ restart = False
             self._check_larger_than_one_attribute(section, key, key_attributes)
 
     def _check_file_attribute(
-        self, section: str, key: str, key_attributes: List[str]
+        self, section: str, key: str, key_attributes: list[str]
     ) -> None:
         """Checks, whether a file specified in key exists.
 
@@ -817,7 +941,7 @@ restart = False
 
         """
         if "file" in key_attributes:
-            file = Path(self.get(section, key))
+            file = pathlib.Path(self.get(section, key))
             if not file.is_file():
                 self.config_errors.append(
                     f"Key {key} in section {section} should point to a file, "
@@ -845,7 +969,7 @@ restart = False
             )
 
     def _check_non_negative_attribute(
-        self, section: str, key: str, key_attributes: List[str]
+        self, section: str, key: str, key_attributes: list[str]
     ) -> None:
         """Checks, whether key is nonnegative.
 
@@ -862,7 +986,7 @@ restart = False
                 )
 
     def _check_positive_attribute(
-        self, section: str, key: str, key_attributes: List[str]
+        self, section: str, key: str, key_attributes: list[str]
     ) -> None:
         """Checks, whether key is positive.
 
@@ -880,7 +1004,7 @@ restart = False
                 )
 
     def _check_less_than_one_attribute(
-        self, section: str, key: str, key_attributes: List[str]
+        self, section: str, key: str, key_attributes: list[str]
     ) -> None:
         """Checks, whether key is less than one.
 
@@ -898,7 +1022,7 @@ restart = False
                 )
 
     def _check_larger_than_one_attribute(
-        self, section: str, key: str, key_attributes: List[str]
+        self, section: str, key: str, key_attributes: list[str]
     ) -> None:
         """Checks, whether key is larger than one.
 
