@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -20,33 +20,35 @@
 from __future__ import annotations
 
 import copy
-from typing import List, Optional, Union
+from typing import TYPE_CHECKING
 
 import fenics
 import numpy as np
 from typing_extensions import Literal
-import ufl
+
+try:
+    import ufl_legacy as ufl
+except ImportError:
+    import ufl
 
 from cashocs import _exceptions
-from cashocs import _loggers
 from cashocs import _utils
+from cashocs import log
+
+if TYPE_CHECKING:
+    from cashocs import _typing
 
 
 class _NewtonSolver:
     """A Newton solver."""
 
-    assembler_shift: fenics.SystemAssembler
-    residual_shift: fenics.PETScVector
-    A_matrix: fenics.PETScMatrix  # pylint: disable=invalid-name
-    b: fenics.PETScVector
-
     def __init__(
         self,
         nonlinear_form: ufl.Form,
         u: fenics.Function,
-        bcs: Union[fenics.DirichletBC, List[fenics.DirichletBC]],
-        derivative: Optional[ufl.Form] = None,
-        shift: Optional[ufl.Form] = None,
+        bcs: fenics.DirichletBC | list[fenics.DirichletBC],
+        derivative: ufl.Form | None = None,
+        shift: ufl.Form | None = None,
         rtol: float = 1e-10,
         atol: float = 1e-10,
         max_iter: int = 50,
@@ -55,10 +57,12 @@ class _NewtonSolver:
         damped: bool = True,
         inexact: bool = True,
         verbose: bool = True,
-        ksp_options: Optional[List[List[Union[str, int, float]]]] = None,
-        A_tensor: Optional[fenics.PETScMatrix] = None,  # pylint: disable=invalid-name
-        b_tensor: Optional[fenics.PETScVector] = None,
+        ksp_options: _typing.KspOption | None = None,
+        A_tensor: fenics.PETScMatrix | None = None,  # pylint: disable=invalid-name
+        b_tensor: fenics.PETScVector | None = None,
         is_linear: bool = False,
+        preconditioner_form: ufl.Form | None = None,
+        linear_solver: _utils.linalg.LinearSolver | None = None,
     ) -> None:
         r"""Initializes self.
 
@@ -96,6 +100,9 @@ class _NewtonSolver:
                 sub-problem.
             is_linear: A boolean flag, which indicates whether the problem is actually
                 linear.
+            preconditioner_form: A UFL form which defines the preconditioner matrix.
+            linear_solver: The linear solver (KSP) which is used to solve the linear
+                systems arising from the discretized PDE
 
         """
         self.nonlinear_form = nonlinear_form
@@ -112,6 +119,15 @@ class _NewtonSolver:
         self.A_tensor = A_tensor  # pylint: disable=invalid-name
         self.b_tensor = b_tensor
         self.is_linear = is_linear
+        if preconditioner_form is not None:
+            if len(preconditioner_form.arguments()) == 1:
+                self.preconditioner_form = fenics.derivative(
+                    preconditioner_form, self.u
+                )
+            else:
+                self.preconditioner_form = preconditioner_form
+        else:
+            self.preconditioner_form = None
 
         self.verbose = verbose if not self.is_linear else False
 
@@ -122,6 +138,8 @@ class _NewtonSolver:
 
         # Setup increment and function for monotonicity test
         self.function_space = u.function_space()
+        self.comm = self.function_space.mesh().mpi_comm()
+
         self.du = fenics.Function(self.function_space)
         self.ddu = fenics.Function(self.function_space)
         self.u_save = fenics.Function(self.function_space)
@@ -133,33 +151,46 @@ class _NewtonSolver:
             self.ksp_options = ksp_options
 
         self.iterations = 0
-        for bc in self.bcs:
-            bc.apply(self.u.vector())
-        # copy the boundary conditions and homogenize them for the increment
-        self.bcs_hom = [fenics.DirichletBC(bc) for bc in self.bcs]
-        for bc in self.bcs_hom:
-            bc.homogenize()
 
         # inexact newton parameters
-        self.eta = 1.0
+        self.eta: float | None = 1.0
         self.eta_max = 0.9999
         self.eta_a = 1.0
         self.gamma = 0.9
         self.lmbd = 1.0
 
         self.assembler = fenics.SystemAssembler(
-            self.derivative, -self.nonlinear_form, self.bcs_hom
+            self.derivative, self.nonlinear_form, self.bcs
         )
         self.assembler.keep_diagonal = True
-        # pylint: disable=invalid-name
-        self.A_fenics = self.A_tensor or fenics.PETScMatrix()
-        self.residual = self.b_tensor or fenics.PETScVector()
 
+        if self.preconditioner_form is not None:
+            self.assembler_pc = fenics.SystemAssembler(
+                self.preconditioner_form, self.nonlinear_form, self.bcs
+            )
+            self.assembler_pc.keep_diagonal = True
+
+        self.comm = self.u.function_space().mesh().mpi_comm()
+        # pylint: disable=invalid-name
+        self.A_fenics = self.A_tensor or fenics.PETScMatrix(self.comm)
+        self.residual = self.b_tensor or fenics.PETScVector(self.comm)
+        self.b = fenics.as_backend_type(self.residual).vec()
+        self.A_matrix = fenics.as_backend_type(self.A_fenics).mat()
+
+        self.P_fenics = fenics.PETScMatrix(self.comm)
+
+        if linear_solver is None:
+            self.linear_solver = _utils.linalg.LinearSolver()
+        else:
+            self.linear_solver = linear_solver
+
+        self.assembler_shift: fenics.SystemAssembler | None = None
+        self.residual_shift: fenics.PETScVector | None = None
         if self.shift is not None:
             self.assembler_shift = fenics.SystemAssembler(
-                self.derivative, self.shift, self.bcs_hom
+                self.derivative, self.shift, self.bcs
             )
-            self.residual_shift = fenics.PETScVector()
+            self.residual_shift = fenics.PETScVector(self.comm)
 
         self.breakdown = False
         self.res = 1.0
@@ -168,26 +199,26 @@ class _NewtonSolver:
 
     def _print_output(self) -> None:
         """Prints the output of the current iteration to the console."""
-        if self.verbose and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
-            prefix = "Newton solver:  "
-
-            if self.iterations % 10 == 0:
-                info_str = (
-                    f"\n{prefix}iter,  "
-                    f"abs. residual (abs. tol),  "
-                    f"rel. residual (rel. tol)\n\n"
-                )
-            else:
-                info_str = ""
-
-            print_str = (
-                f"{prefix}{self.iterations:4d},  "
-                f"{self.res:>13.3e} ({self.atol:.2e}),  "
-                f"{self.res/self.res_0:>13.3e} ({self.rtol:.2e})"
+        if self.iterations % 10 == 0:
+            info_str = (
+                "\niter,  abs. residual (abs. tol),  rel. residual (rel. tol)\n\n"
             )
+        else:
+            info_str = ""
 
-            print(info_str + print_str, flush=True)
+        print_str = (
+            f"{self.iterations:4d},  "
+            f"{self.res:>13.3e} ({self.atol:.2e}),  "
+            f"{self.res / self.res_0:>13.3e} ({self.rtol:.2e})"
+        )
+        if self.verbose:
+            if self.comm.rank == 0:
+                print(info_str + print_str, flush=True)
+            self.comm.barrier()
+        else:
+            log.debug(info_str + print_str)
 
+    @log.profile_execution_time("assembling the Jacobian for Newton's method")
     def _assemble_matrix(self) -> None:
         """Assembles the matrix for solving the linear problem."""
         self.assembler.assemble(self.A_fenics)
@@ -196,28 +227,37 @@ class _NewtonSolver:
             self.A_fenics
         ).mat()
 
+        if self.preconditioner_form is not None:
+            self.assembler_pc.assemble(self.P_fenics)
+            self.P_fenics.ident_zeros()
+            self.P_matrix = fenics.as_backend_type(  # pylint: disable=invalid-name
+                self.P_fenics
+            ).mat()
+        else:
+            self.P_matrix = None
+
     def _compute_eta_inexact(self) -> None:
         """Computes the parameter ``eta`` for the inexact Newton method."""
-        if self.inexact:
+        if self.inexact and isinstance(self.eta, float):
             if self.iterations == 1:
-                self.eta = self.eta_max
+                eta_new = self.eta_max
             elif self.gamma * pow(self.eta, 2) <= 0.1:
-                self.eta = np.minimum(self.eta_max, self.eta_a)
+                eta_new = np.minimum(self.eta_max, self.eta_a)
             else:
-                self.eta = np.minimum(
+                eta_new = np.minimum(
                     self.eta_max,
                     np.maximum(self.eta_a, self.gamma * pow(self.eta, 2)),
                 )
 
             self.eta = np.minimum(
                 self.eta_max,
-                np.maximum(self.eta, 0.5 * self.tol / self.res),
+                np.maximum(eta_new, 0.5 * self.tol / self.res),
             )
         else:
-            self.eta = self.rtol * 1e-1
+            self.eta = None
 
         if self.is_linear:
-            self.eta = self.rtol * 1e-1
+            self.eta = None
 
     def _check_for_nan_residual(self) -> None:
         """Checks, whether the residual is nan. If yes, raise a NotConvergedError."""
@@ -231,12 +271,20 @@ class _NewtonSolver:
             A solution of the nonlinear problem.
 
         """
+        log.begin(
+            "Solving the nonlinear PDE system with Newton's method.", level=log.DEBUG
+        )
         self._compute_residual()
 
         self.res_0 = self.residual.norm(self.norm_type)
         if self.res_0 == 0.0:  # pragma: no cover
-            if self.verbose and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
-                print("Residual vanishes, input is already a solution.", flush=True)
+            message = "Residual vanishes, input is already a solution."
+            if self.verbose:
+                if self.comm.rank == 0:
+                    print(message, flush=True)
+                self.comm.barrier()
+            else:
+                log.debug(message)
             return self.u
 
         self.res = self.res_0
@@ -256,27 +304,24 @@ class _NewtonSolver:
             self.u_save.vector().apply("")
 
             self._compute_eta_inexact()
-            _utils.solve_linear_problem(
+            self.linear_solver.solve(
+                self.du,
                 A=self.A_matrix,
                 b=self.b,
-                x=self.du.vector().vec(),
                 ksp_options=self.ksp_options,
                 rtol=self.eta,
                 atol=self.atol / 10.0,
+                P=self.P_matrix,
             )
-            self.du.vector().apply("")
 
             if self.is_linear:
-                self.u.vector().vec().axpy(1.0, self.du.vector().vec())
+                self.u.vector().vec().axpy(-1.0, self.du.vector().vec())
                 self.u.vector().apply("")
                 break
 
             self._backtracking_line_search()
             self._compute_residual()
 
-            for bc in self.bcs_hom:
-                bc.apply(self.residual)
-
             res_prev = self.res
             self.res = self.residual.norm(self.norm_type)
             self._check_for_nan_residual()
@@ -286,6 +331,7 @@ class _NewtonSolver:
             if self._check_for_convergence():
                 break
 
+        log.end()
         self._check_if_successful()
 
         return self.u
@@ -293,12 +339,15 @@ class _NewtonSolver:
     def _check_for_convergence(self) -> bool:
         """Checks, whether the desired convergence tolerance has been reached."""
         if self.res <= self.tol:
-            if self.verbose and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
-                print(
-                    f"\nNewton Solver converged "
-                    f"after {self.iterations:d} iterations.\n",
-                    flush=True,
-                )
+            convergence_message = (
+                f"Newton Solver converged after {self.iterations:d} iterations."
+            )
+            if self.verbose:
+                if self.comm.rank == 0:
+                    print(convergence_message, flush=True)
+                self.comm.barrier()
+            else:
+                log.debug(convergence_message)
             return True
 
         else:
@@ -326,12 +375,18 @@ class _NewtonSolver:
                 "Maximum number of iterations were exceeded.",
             )
 
+    @log.profile_execution_time("assembling the residual for Newton's method")
     def _compute_residual(self) -> None:
         """Computes the residual of the nonlinear system."""
-        self.assembler.assemble(self.residual)
-        if self.shift is not None:
-            self.assembler_shift.assemble(self.residual_shift)
-            self.residual[:] += self.residual_shift[:]
+        self.residual = fenics.PETScVector(self.comm)
+        self.assembler.assemble(self.residual, self.u.vector())
+        if (
+            self.shift is not None
+            and self.assembler_shift is not None
+            and self.residual_shift is not None
+        ):
+            self.assembler_shift.assemble(self.residual_shift, self.u.vector())
+            self.residual[:] -= self.residual_shift[:]
 
         self.b = fenics.as_backend_type(self.residual).vec()
 
@@ -353,18 +408,18 @@ class _NewtonSolver:
         """Performs a backtracking line search for the damped Newton method."""
         if self.damped:
             while True:
-                self.u.vector().vec().axpy(self.lmbd, self.du.vector().vec())
+                self.u.vector().vec().axpy(-self.lmbd, self.du.vector().vec())
                 self.u.vector().apply("")
                 self._compute_residual()
-                _utils.solve_linear_problem(
+                self.linear_solver.solve(
+                    self.ddu,
                     A=self.A_matrix,
                     b=self.b,
-                    x=self.ddu.vector().vec(),
                     ksp_options=self.ksp_options,
                     rtol=self.eta,
                     atol=self.atol / 10.0,
+                    P=self.P_matrix,
                 )
-                self.ddu.vector().apply("")
 
                 if (
                     self.ddu.vector().norm(self.norm_type)
@@ -381,16 +436,16 @@ class _NewtonSolver:
                     self.breakdown = True
                     break
         else:
-            self.u.vector().vec().axpy(1.0, self.du.vector().vec())
+            self.u.vector().vec().axpy(-1.0, self.du.vector().vec())
             self.u.vector().apply("")
 
 
 def newton_solve(
     nonlinear_form: ufl.Form,
     u: fenics.Function,
-    bcs: Union[fenics.DirichletBC, List[fenics.DirichletBC]],
-    derivative: Optional[ufl.Form] = None,
-    shift: Optional[ufl.Form] = None,
+    bcs: fenics.DirichletBC | list[fenics.DirichletBC],
+    derivative: ufl.Form | None = None,
+    shift: ufl.Form | None = None,
     rtol: float = 1e-10,
     atol: float = 1e-10,
     max_iter: int = 50,
@@ -399,10 +454,12 @@ def newton_solve(
     damped: bool = True,
     inexact: bool = True,
     verbose: bool = True,
-    ksp_options: Optional[List[List[Union[str, int, float]]]] = None,
-    A_tensor: Optional[fenics.PETScMatrix] = None,  # pylint: disable=invalid-name
-    b_tensor: Optional[fenics.PETScVector] = None,
+    ksp_options: _typing.KspOption | None = None,
+    A_tensor: fenics.PETScMatrix | None = None,  # pylint: disable=invalid-name
+    b_tensor: fenics.PETScVector | None = None,
     is_linear: bool = False,
+    preconditioner_form: ufl.Form | None = None,
+    linear_solver: _utils.linalg.LinearSolver | None = None,
 ) -> fenics.Function:
     r"""Solves a nonlinear problem with Newton\'s method.
 
@@ -438,6 +495,9 @@ def newton_solve(
             sub-problem.
         is_linear: A boolean flag, which indicates whether the problem is actually
             linear.
+        preconditioner_form: A UFL form which defines the preconditioner matrix.
+        linear_solver: The linear solver (KSP) which is used to solve the linear
+            systems arising from the discretized PDE.
 
     Returns:
         The solution of the nonlinear variational problem, if converged. This overwrites
@@ -485,100 +545,9 @@ def newton_solve(
         A_tensor=A_tensor,
         b_tensor=b_tensor,
         is_linear=is_linear,
+        preconditioner_form=preconditioner_form,
+        linear_solver=linear_solver,
     )
 
     solution = solver.solve()
     return solution
-
-
-def damped_newton_solve(
-    nonlinear_form: ufl.Form,
-    u: fenics.Function,
-    bcs: Union[fenics.DirichletBC, List[fenics.DirichletBC]],
-    derivative: Optional[ufl.Form] = None,
-    rtol: float = 1e-10,
-    atol: float = 1e-10,
-    max_iter: int = 50,
-    convergence_type: Literal["combined", "rel", "abs"] = "combined",
-    norm_type: Literal["l2", "linf"] = "l2",
-    damped: bool = True,
-    verbose: bool = True,
-    ksp_options: Optional[List[List[Union[str, int, float]]]] = None,
-) -> fenics.Function:  # pragma: no cover
-    r"""Damped Newton solve interface, only here for compatibility reasons.
-
-    Args:
-        nonlinear_form: The variational form of the nonlinear problem to be solved by
-            Newton's method.
-        u: The sought solution / initial guess. It is not assumed that the initial guess
-            satisfies the Dirichlet boundary conditions, they are applied automatically.
-            The method overwrites / updates this Function.
-        bcs: A list of DirichletBCs for the nonlinear variational problem.
-        derivative: The Jacobian of nonlinear_form, used for the Newton method. Default
-            is None, and in this case the Jacobian is computed automatically with AD.
-        rtol: Relative tolerance of the solver if convergence_type is either
-            ``'combined'`` or ``'rel'`` (default is ``rtol = 1e-10``).
-        atol: Absolute tolerance of the solver if convergence_type is either
-            ``'combined'`` or ``'abs'`` (default is ``atol = 1e-10``).
-        max_iter: Maximum number of iterations carried out by the method (default is
-            ``max_iter = 50``).
-        convergence_type: Determines the type of stopping criterion that is used.
-        norm_type: Determines which norm is used in the stopping criterion.
-        damped: If ``True``, then a damping strategy is used. If ``False``, the
-            classical Newton-Raphson iteration (without damping) is used (default is
-            ``True``).
-        verbose: If ``True``, prints status of the iteration to the console (default is
-            ``True``).
-        ksp_options: The list of options for the linear solver.
-
-    Returns:
-        The solution of the nonlinear variational problem, if converged. This overwrites
-        the input function u.
-
-    Examples:
-        Consider the problem
-
-        .. math::
-            \begin{alignedat}{2}
-            - \Delta u + u^3 &= 1 \quad &&\text{ in } \Omega=(0,1)^2 \\
-            u &= 0 \quad &&\text{ on } \Gamma.
-            \end{alignedat}
-
-        This is solved with the code ::
-
-            from fenics import *
-            import cashocs
-
-            mesh, _, boundaries, dx, _, _ = cashocs.regular_mesh(25)
-            V = FunctionSpace(mesh, 'CG', 1)
-
-            u = Function(function_space)
-            v = TestFunction(function_space)
-            F = inner(grad(u), grad(v))*dx + pow(u,3)*v*dx - Constant(1)*v*dx
-            bcs = cashocs.create_dirichlet_bcs(V, Constant(0.0), boundaries, [1,2,3,4])
-            cashocs.newton_solve(F, u, bcs)
-
-    .. deprecated:: 1.5.0
-        This is replaced by cashocs.newton_solve and will be removed in the future.
-
-    """
-    _loggers.warning(
-        "DEPREACTION WARNING: cashocs.damped_newton_solve is replaced "
-        "by cashocs.newton_solve and will be removed in the future."
-    )
-
-    return newton_solve(
-        nonlinear_form,
-        u,
-        bcs,
-        derivative=derivative,
-        shift=None,
-        rtol=rtol,
-        atol=atol,
-        max_iter=max_iter,
-        convergence_type=convergence_type,
-        norm_type=norm_type,
-        damped=damped,
-        verbose=verbose,
-        ksp_options=ksp_options,
-    )
