@@ -1,4 +1,4 @@
-# Copyright (C) 2020-2022 Sebastian Blauth
+# Copyright (C) 2020-2025 Fraunhofer ITWM and Sebastian Blauth
 #
 # This file is part of cashocs.
 #
@@ -15,9 +15,9 @@
 # You should have received a copy of the GNU General Public License
 # along with cashocs.  If not, see <https://www.gnu.org/licenses/>.
 
-"""The space_mapping module contains utilities for using the space mapping technique."""
+"""Space mapping algorithms."""
 
-from cashocs.space_mapping import optimal_control  # noqa: F401
-from cashocs.space_mapping import shape_optimization  # noqa: F401
+from cashocs.space_mapping import optimal_control
+from cashocs.space_mapping import shape_optimization
 
 __all__ = ["optimal_control", "shape_optimization"]
